"""
This file contains the tests for periodic tasks.

In particular, this tests
lib/periodictask.py
"""
from datetime import datetime, timedelta

from dateutil.parser import parse as parse_timestamp
from dateutil.tz import gettz, tzutc
from mock import mock

from privacyidea.lib.error import ServerError, ParameterError, ResourceNotFoundError
from privacyidea.lib.periodictask import calculate_next_timestamp, set_periodic_task, get_periodic_tasks, \
    enable_periodic_task, delete_periodic_task, set_periodic_task_last_run, get_scheduled_periodic_tasks, \
    get_periodic_task_by_name, TASK_MODULES, execute_task, get_periodic_task_by_id
from privacyidea.lib.task.base import BaseTask
from privacyidea.models import PeriodicTask
from .base import MyTestCase


class BasePeriodicTaskTestCase(MyTestCase):
    def test_01_calculate_next_timestamp_utc(self):
        # The easy case: calculate everything in UTC
        tzinfo = tzutc()

        # every day at 08:00
        task1 = {
            "id": 1,
            "active": True,
            "name": "task one",
            "interval": "0 8 * * *",
            "last_update": parse_timestamp("2018-06-23 07:55:00 UTC"),
            "nodes": ["foo", "bar", "baz"],
            "taskmodule": "some.module",
            "options": {"KEY2": "value number 2",
                        "key 4": "1234"},
            "last_runs": {
                "foo": parse_timestamp("2018-06-25 08:04:30 UTC"),
                "bar": parse_timestamp("2018-06-24 07:05:37 UTC"),
            }
        }

        self.assertEqual(calculate_next_timestamp(task1, "foo", tzinfo),
                         parse_timestamp("2018-06-26 08:00 UTC"))
        self.assertEqual(calculate_next_timestamp(task1, "bar", tzinfo),
                         parse_timestamp("2018-06-24 08:00 UTC"))
        # the next run of baz is calculated based on last_update
        self.assertEqual(calculate_next_timestamp(task1, "baz", tzinfo),
                         parse_timestamp("2018-06-23 08:00 UTC"))

        # no last run recorded
        task1b = {
            "id": 1,
            "active": True,
            "name": "task one",
            "interval": "0 8 * * *",
            "last_update": parse_timestamp("2018-06-24 07:55:00 UTC"),
            "nodes": ["foo", "bar"],
            "taskmodule": "some.module",
            "options": {"KEY2": "value number 2",
                        "key 4": "1234"},
            "last_runs": {}
        }

        self.assertEqual(calculate_next_timestamp(task1b, "foo", tzinfo),
                         parse_timestamp("2018-06-24 08:00 UTC"))
        self.assertEqual(calculate_next_timestamp(task1b, "bar", tzinfo),
                         parse_timestamp("2018-06-24 08:00 UTC"))

        # now, "foo" has a last run!
        task1b["last_runs"]["foo"] = parse_timestamp("2018-06-24 08:00 UTC")
        self.assertEqual(calculate_next_timestamp(task1b, "foo", tzinfo),
                         parse_timestamp("2018-06-25 08:00 UTC"))
        # ... bar has still not run
        self.assertEqual(calculate_next_timestamp(task1b, "bar", tzinfo),
                         parse_timestamp("2018-06-24 08:00 UTC"))

        # every weekday
        task2 = {
            "id": 2,
            "active": True,
            "name": "task two",
            "interval": "0 0 * * 1-5",
            "last_update": parse_timestamp("2018-06-24 08:00:00 UTC"),
            "nodes": ["foo", "bar"],
            "taskmodule": "some.module",
            "options": {"KEY2": "value number 2",
                        "key 4": "1234"},
            "last_runs": {
                "localhost": parse_timestamp("2018-06-29 00:00 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task2, "localhost", tzinfo),
                         parse_timestamp("2018-07-02 00:00 UTC"))

        # at 00:05 in August
        task3 = {
            "id": 3,
            "active": True,
            "name": "task two",
            "interval": "5 0 * 8 *",
            "last_update": parse_timestamp("2018-06-24 08:00:00 UTC"),
            "nodes": ["foo", "bar"],
            "taskmodule": "some.module",
            "options": {"KEY2": "value number 2",
                        "key 4": "1234"},
            "last_runs": {
                "localhost": parse_timestamp("2017-08-31 00:06 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task3, "localhost", tzinfo),
                         parse_timestamp("2018-08-01 00:05 UTC"))

        # malformed
        task4 = {
            "id": 3,
            "active": True,
            "name": "task two",
            "interval": "every two days",
            "last_update": parse_timestamp("2018-06-24 08:00:00 UTC"),
            "nodes": ["foo", "bar"],
            "taskmodule": "some.module",
            "options": {"KEY2": "value number 2",
                        "key 4": "1234"},
            "last_runs": {
                "localhost": parse_timestamp("2017-08-31 00:06 UTC"),
            }
        }
        with self.assertRaises(ValueError):
            calculate_next_timestamp(task4, "localhost")

    def test_02_calculate_next_timestamp_localtime(self):
        # The harder case: Calculate everything in a local timezone
        # There is no DST in russia, so we operate in +03:00
        tzinfo = gettz("Europe/Moscow")

        # every day at 08:00
        task = {
            "interval": "0 8 * * *",
            "last_update": parse_timestamp("2018-06-24 02:30 UTC"),
            "last_runs": {
                "foo": parse_timestamp("2018-06-25 05:04:30 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task, "foo", tzinfo),
                         parse_timestamp("2018-06-26 05:00 UTC"))
        self.assertEqual(calculate_next_timestamp(task, "bar", tzinfo),
                         parse_timestamp("2018-06-24 05:00 UTC"))
        self.assertEqual(calculate_next_timestamp(task, "this_node_does_not_exist", tzinfo),
                         parse_timestamp("2018-06-24 05:00 UTC"))

        # every day at 08:00
        task = {
            "interval": "0 8 * * *",
            "last_update": parse_timestamp("2018-06-24 02:30 UTC"),
            "last_runs": {
                "foo": parse_timestamp("2018-06-25 04:04:30 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task, "foo", tzinfo),
                         parse_timestamp("2018-06-25 05:00 UTC"))

        # every day at midnight
        task = {
            "interval": "0 0 * * *",
            "last_update": parse_timestamp("2018-06-24 02:30 UTC"),
            "last_runs": {
                "foo": parse_timestamp("2018-06-25 21:01 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task, "foo", tzinfo),
                         parse_timestamp("2018-06-26 21:00 UTC"))

        # every wednesday at midnight
        task = {
            "interval": "0 0 * * 3",
            "last_update": parse_timestamp("2018-06-24 02:30 UTC"),
            "last_runs": {
                "foo": parse_timestamp("2018-06-24 21:00 UTC"),  # this is actually monday 00:00 in russia
            }
        }
        self.assertEqual(calculate_next_timestamp(task, "foo", tzinfo),
                         parse_timestamp("2018-06-26 21:00 UTC"))

        # every 15th at 01:00
        task = {
            "interval": "0 1 15 * *",
            "last_update": parse_timestamp("2018-06-24 02:30 UTC"),
            "last_runs": {
                "foo": parse_timestamp("2018-05-15 00:00 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task, "foo", tzinfo),
                         parse_timestamp("2018-06-14 22:00 UTC"))

        # every 15th at 01:00
        task = {
            "interval": "0 1 15 * *",
            "last_update": parse_timestamp("2018-06-24 02:30 UTC"),
            "last_runs": {
                "foo": parse_timestamp("2018-05-14 21:59 UTC"),
            }
        }
        self.assertEqual(calculate_next_timestamp(task, "foo", tzinfo),
                         parse_timestamp("2018-05-14 22:00 UTC"))

    def test_03_crud(self):
        task1 = set_periodic_task("task one", "0 0 1 * *", ["pinode1"], "some.task.module", 3, {
            "key1": 1,
            "key2": False
        }, retry_if_failed=False)
        task2 = set_periodic_task("task two", "0 0 * * WED", ["pinode2"], "some.task.module", 1, {
            "key1": "value",
            "key2": "foo"
        }, active=False, retry_if_failed=True)
        task3 = set_periodic_task("task three", "30 * * * *", ["pinode1", "pinode2"], "some.task.module", 2, {
            "key1": 1234,
            "key2": 5678,
        })

        with self.assertRaises(ParameterError):
            set_periodic_task("task four", "61 * * * *", ["pinode1", "pinode2"], "some.task.module", 1, {
                "key1": 1234,
                "key2": 5678,
            })
        with self.assertRaises(ParameterError):
            set_periodic_task("task four", "1 * * * *", ["pinode1", "pinode2"], "some.task.module", -3, {
                "key1": 1234,
                "key2": 5678,
            })

        task1_last_update = get_periodic_task_by_id(task1)["last_update"]

        self.assertEqual(get_periodic_task_by_name("task three")["id"], task3)
        with self.assertRaises(ResourceNotFoundError):
            get_periodic_task_by_name("task does not exist")

        self.assertEqual(get_periodic_task_by_id(task3)["name"], "task three")
        with self.assertRaises(ResourceNotFoundError):
            get_periodic_task_by_id(1337)

        self.assertEqual(len(PeriodicTask.query.all()), 3)

        # Updating an nonexistent task fails
        with self.assertRaises(ResourceNotFoundError):
            set_periodic_task("some task", "0 0 1 * *", ["pinode1"], "some.task.module", 5, {}, id=123456)

        task1_modified = set_periodic_task("every month", "0 0 1 * *", ["pinode1"], "some.task.module", 3, {
            "key1": 123,
            "key3": True,
        }, id=task1)

        self.assertEqual(len(PeriodicTask.query.all()), 3)
        self.assertEqual(task1, task1_modified)
        # we have updated the task definition
        self.assertGreater(get_periodic_task_by_id(task1)["last_update"], task1_last_update)
        self.assertEqual(get_periodic_tasks(name="every month")[0]["options"],
                         {"key1": "123", "key3": "True"})

        all_tasks = get_periodic_tasks()
        self.assertEqual(len(all_tasks), 3)
        # ordered by ordering
        self.assertEqual([task["name"] for task in all_tasks], ["task two", "task three", "every month"])

        active_tasks = get_periodic_tasks(active=True)
        self.assertEqual(len(active_tasks), 2)
        self.assertEqual([task["name"] for task in active_tasks], ["task three", "every month"])

        active_tasks_on_pinode2 = get_periodic_tasks(active=True, node="pinode2")
        self.assertEqual(len(active_tasks_on_pinode2), 1)
        self.assertEqual(active_tasks_on_pinode2[0]["name"], "task three")

        enable_periodic_task(task2)

        active_tasks_on_pinode2 = get_periodic_tasks(active=True, node="pinode2")
        self.assertEqual(len(active_tasks_on_pinode2), 2)
        self.assertEqual([task["name"] for task in active_tasks_on_pinode2],
                         ["task two", "task three"])

        active_tasks_on_pinode1 = get_periodic_tasks(active=True, node="pinode1")
        self.assertEqual(len(active_tasks_on_pinode1), 2)
        self.assertEqual([task["name"] for task in active_tasks_on_pinode1],
                         ["task three", "every month"])

        active_tasks_on_pinode3 = get_periodic_tasks(active=True, node="pinode3")
        self.assertEqual(active_tasks_on_pinode3, [])

        enable_periodic_task(task1, False)
        delete_periodic_task(task3)
        with self.assertRaises(ResourceNotFoundError):
            enable_periodic_task(task3)

        active_tasks_on_pinode1 = get_periodic_tasks(active=True, node="pinode1")
        self.assertEqual(active_tasks_on_pinode1, [])

        tasks_on_pinode1 = get_periodic_tasks(node="pinode1")
        self.assertEqual(len(tasks_on_pinode1), 1)
        self.assertEqual(tasks_on_pinode1[0]["name"], "every month")

        delete_periodic_task(task1)
        delete_periodic_task(task2)
        with self.assertRaises(ResourceNotFoundError):
            delete_periodic_task(task2)

        self.assertEqual(get_periodic_tasks(), [])

    def test_04_last_run(self):
        task1_id = set_periodic_task("task one", "*/5 * * * *", ["pinode1", "pinode2"], "some.task.module", 3, {
            "key1": 1,
            "key2": False
        })
        self.assertEqual(len(PeriodicTask.query.all()), 1)
        task1_entry = PeriodicTask.query.filter_by(id=task1_id).one()

        # We have no initial last runs
        self.assertEqual(len(list(task1_entry.last_runs)), 0)

        set_periodic_task_last_run(task1_id, "pinode1", parse_timestamp("2018-06-26 08:00+02:00"))
        set_periodic_task_last_run(task1_id, "pinode1", parse_timestamp("2018-06-26 08:05+02:00"))

        task1 = get_periodic_tasks("task one")[0]
        self.assertEqual(len(list(task1_entry.last_runs)), 1)
        self.assertEqual(task1_entry.last_runs[0].timestamp,
                         parse_timestamp("2018-06-26 06:05"))
        self.assertEqual(task1["last_runs"]["pinode1"],
                         parse_timestamp("2018-06-26 06:05 UTC"))

        set_periodic_task_last_run(task1_id, "pinode2", parse_timestamp("2018-06-26 08:10+01:00"))
        set_periodic_task_last_run(task1_id, "pinode3", parse_timestamp("2018-06-26 08:10-08:00"))
        task1 = get_periodic_tasks("task one")[0]
        self.assertEqual(task1["last_runs"]["pinode1"],
                         parse_timestamp("2018-06-26 06:05 UTC"))
        self.assertEqual(task1["last_runs"]["pinode2"],
                         parse_timestamp("2018-06-26 07:10 UTC"))
        self.assertEqual(task1["last_runs"]["pinode3"],
                         parse_timestamp("2018-06-26 16:10 UTC"))

        delete_periodic_task(task1_id)

    def test_05_scheduling(self):
        # this unit test operates in russian time
        tzinfo = gettz("Europe/Moscow")

        # at midnight on each 1st
        task1 = set_periodic_task("task one", "0 0 1 * *", ["pinode1"], "some.task.module", 3, {
            "key1": 1,
            "key2": False
        })
        # at 08:00 on wednesdays
        current_utc_time = parse_timestamp("2018-05-31 05:08:00")
        with mock.patch('privacyidea.models.datetime') as mock_dt:
            mock_dt.utcnow.return_value = current_utc_time
            task2 = set_periodic_task("task two", "0 8 * * WED", ["pinode2", "pinode3"], "some.task.module", 1, {
                "key1": "value",
                "key2": "foo"
            }, active=False)
        self.assertEqual(get_periodic_task_by_id(task2)["last_update"],
                         parse_timestamp("2018-05-31 08:08:00+03:00"))
        self.assertEqual(get_periodic_task_by_id(task2)["last_runs"], {})

        # every 30 minutes, on Tuesdays
        task3 = set_periodic_task("task three", "*/30 * * * 2", ["pinode1", "pinode2"], "some.task.module", 2, {
            "key1": 1234,
            "key2": 5678,
        })
        # on each 1st of august at midnight
        task4 = set_periodic_task("task four", "0 0 1 8 *", ["pinode2"], "some.task.module", 0)

        # we need some last runs
        set_periodic_task_last_run(task1, "pinode1", parse_timestamp("2018-06-01 00:00:05+03:00"))

        # no last run for pinode3 here!
        set_periodic_task_last_run(task2, "pinode2", parse_timestamp("2018-06-20 08:00:05+03:00"))

        set_periodic_task_last_run(task3, "pinode1", parse_timestamp("2018-06-26 11:36:37+03:00"))
        set_periodic_task_last_run(task3, "pinode2", parse_timestamp("2018-06-26 11:30:33+03:00"))

        set_periodic_task_last_run(task4, "pinode2", parse_timestamp("2017-08-01 00:00:43+03:00"))

        self.assertEqual([task["name"] for task in get_periodic_tasks()],
                         ["task four", "task two", "task three", "task one"])

        # Invalid timestamp
        with self.assertRaises(ParameterError):
            get_scheduled_periodic_tasks("pinode1", parse_timestamp("2017-08-01 00:00:00"), tzinfo)

        # On pinode1:
        # task1 at midnight on each 1st
        # task3 every 30 minutes on tuesdays

        # On pinode2:
        # task2 on 08:00 on wednesdays, but it is inactive
        # task3 every 30 minutes on tuesdays
        # task4 on each 1st August at midnight

        # 26th June (Tuesday), 11:59
        # No tasks on both nodes
        current_timestamp = parse_timestamp("2018-06-26 11:59+03:00")

        scheduled = get_scheduled_periodic_tasks("pinode1", current_timestamp, tzinfo)
        self.assertEqual(scheduled, [])

        scheduled = get_scheduled_periodic_tasks("pinode2", current_timestamp, tzinfo)
        self.assertEqual(scheduled, [])

        # 26th June (Tuesday), 12:00
        # Run task3 on both nodes
        current_timestamp = parse_timestamp("2018-06-26 12:00+03:00")

        scheduled = get_scheduled_periodic_tasks("pinode1", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task three"])

        scheduled = get_scheduled_periodic_tasks("pinode2", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task three"])

        # 1th August (Wednesday), 13:57
        # Assume task3 has been run successfully on 30th July (Tuesday)
        set_periodic_task_last_run(task3, "pinode1", parse_timestamp("2018-08-01 00:00+03:00"))
        set_periodic_task_last_run(task3, "pinode2", parse_timestamp("2018-08-01 00:00+03:00"))

        # On pinode1, run task1
        # On pinode2, run task4
        current_timestamp = parse_timestamp("2018-08-01 11:59+03:00")

        scheduled = get_scheduled_periodic_tasks("pinode1", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task one"])

        scheduled = get_scheduled_periodic_tasks("pinode2", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task four"])

        # Enable task2, now we also have to run it on pinode2 and pinode3
        with mock.patch('privacyidea.models.datetime') as mock_dt:
            mock_dt.utcnow.return_value = current_utc_time
            enable_periodic_task(task2)

        scheduled = get_scheduled_periodic_tasks("pinode1", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task one"])

        scheduled = get_scheduled_periodic_tasks("pinode2", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task four", "task two"])

        scheduled = get_scheduled_periodic_tasks("pinode3", current_timestamp, tzinfo)
        self.assertEqual([task["name"] for task in scheduled], ["task two"])

        # Simulate runs
        set_periodic_task_last_run(task1, "pinode1", current_timestamp)
        set_periodic_task_last_run(task2, "pinode2", current_timestamp)
        set_periodic_task_last_run(task2, "pinode3", current_timestamp)
        set_periodic_task_last_run(task4, "pinode2", current_timestamp)

        # Now, we don't have to run anything
        current_timestamp += timedelta(seconds=1)

        self.assertEqual(get_scheduled_periodic_tasks("pinode1", current_timestamp, tzinfo), [])
        self.assertEqual(get_scheduled_periodic_tasks("pinode2", current_timestamp, tzinfo), [])

        delete_periodic_task(task1)
        delete_periodic_task(task2)
        delete_periodic_task(task3)
        delete_periodic_task(task4)

    def test_06_execute_task(self):
        assertEqual = self.assertEqual

        class _TestTask(BaseTask):
            identifier = "Test"
            description = "foo"

            def do(self, params):
                assertEqual(params["key"], "value")
                return True

        with self.assertRaises(ParameterError):
            execute_task("Test", {"key": "value"})

        with mock.patch.dict(TASK_MODULES, values={"Test": _TestTask}):
            ret = execute_task("Test", {"key": "value"})
            self.assertTrue(ret)



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
contains Errors and Exceptions
"""

import logging

log = logging.getLogger(__name__)


class ERROR:
    SUBSCRIPTION = 101
    TOKENADMIN = 301
    CONFIGADMIN = 302
    POLICY = 303
    IMPORTADMIN = 304
    VALIDATE = 401
    REGISTRATION = 402
    AUTHENTICATE = 403
    AUTHENTICATE_WRONG_CREDENTIALS = 4031
    AUTHENTICATE_MISSING_USERNAME = 4032
    AUTHENTICATE_AUTH_HEADER = 4033
    AUTHENTICATE_DECODING_ERROR = 4304
    AUTHENTICATE_TOKEN_EXPIRED = 4305
    AUTHENTICATE_MISSING_RIGHT = 4306
    AUTHENTICATE_ILLEGAL_METHOD = 4307
    ENROLLMENT = 404
    CA = 503
    CA_CSR_INVALID = 504
    CA_CSR_PENDING = 505
    RESOURCE_NOT_FOUND = 601
    HSM = 707
    SELFSERVICE = 807
    DATABASE = 902
    SERVER = 903
    USER = 904
    PARAMETER = 905
    RESOLVER = 907
    PARAMETER_USER_MISSING = 9051
    CONTAINER = 3000
    CONTAINER_NOT_REGISTERED = 3001
    CONTAINER_INVALID_CHALLENGE = 3002
    CONTAINER_ROLLOVER = 3003


class privacyIDEAError(Exception):

    def __init__(self, description="privacyIDEAError!", id=10):
        self.id = id
        self.message = description
        Exception.__init__(self, description)

    def getId(self):
        return self.id

    def getDescription(self):
        return self.message

    def __str__(self):
        if isinstance(self.message, str):
            return f"ERR{self.id}: {self.message}"

        return f"ERR{self.id}: {self.message!r}"

    def __repr__(self):
        return f"{type(self).__name__}(description={self.message!r}, id={self.id:d})"


class SubscriptionError(privacyIDEAError):
    def __init__(self, description=None, application=None, id=ERROR.SUBSCRIPTION):
        self.id = id
        self.message = description
        self.application = application
        privacyIDEAError.__init__(self, description, id=self.id)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        ret = '{0!s}({1!r}, application={2!s})'.format(type(
            self).__name__, self.message, self.application)
        return ret


class TokenImportException(privacyIDEAError):
    def __init__(self, description, id=ERROR.IMPORTADMIN):
        privacyIDEAError.__init__(self, description=description, id=id)


class AuthError(privacyIDEAError):
    def __init__(self, description, id=ERROR.AUTHENTICATE, details=None):
        self.details = details
        privacyIDEAError.__init__(self, description=description, id=id)


class ResourceNotFoundError(privacyIDEAError):
    def __init__(self, description, id=ERROR.RESOURCE_NOT_FOUND):
        privacyIDEAError.__init__(self, description=description, id=id)


class PolicyError(privacyIDEAError):
    def __init__(self, description, id=ERROR.POLICY):
        privacyIDEAError.__init__(self, description=description, id=id)


class ValidateError(privacyIDEAError):
    def __init__(self, description="validation error!", id=ERROR.VALIDATE):
        privacyIDEAError.__init__(self, description=description, id=id)


class RegistrationError(privacyIDEAError):
    def __init__(self, description="registration error!", id=ERROR.REGISTRATION):
        privacyIDEAError.__init__(self, description=description, id=id)


class EnrollmentError(privacyIDEAError):
    def __init__(self, description="enrollment error!", id=ERROR.ENROLLMENT):
        privacyIDEAError.__init__(self, description=description, id=id)


class TokenAdminError(privacyIDEAError):
    def __init__(self, description="token admin error!", id=ERROR.TOKENADMIN):
        privacyIDEAError.__init__(self, description=description, id=id)


class ConfigAdminError(privacyIDEAError):
    def __init__(self, description="config admin error!", id=ERROR.CONFIGADMIN):
        privacyIDEAError.__init__(self, description=description, id=id)


class CAError(privacyIDEAError):
    def __init__(self, description="CA error!", id=ERROR.CA):
        privacyIDEAError.__init__(self, description=description, id=id)


class CSRError(CAError):
    def __init__(self, description="CSR invalid", id=ERROR.CA_CSR_INVALID):
        privacyIDEAError.__init__(self, description=description, id=id)


class CSRPending(CAError):
    def __init__(self, description="CSR pending", id=ERROR.CA_CSR_PENDING, requestId=None):
        privacyIDEAError.__init__(self, description=description, id=id)
        self.requestId = requestId


class UserError(privacyIDEAError):
    def __init__(self, description="user error!", id=ERROR.USER):
        privacyIDEAError.__init__(self, description=description, id=id)


class ServerError(privacyIDEAError):
    def __init__(self, description="server error!", id=ERROR.SERVER):
        privacyIDEAError.__init__(self, description=description, id=id)


class HSMException(privacyIDEAError):
    def __init__(self, description="hsm error!", id=ERROR.HSM):
        privacyIDEAError.__init__(self, description=description, id=id)


class SelfserviceException(privacyIDEAError):
    def __init__(self, description="selfservice error!", id=ERROR.SELFSERVICE):
        privacyIDEAError.__init__(self, description=description, id=id)


class ParameterError(privacyIDEAError):

    def __init__(self, description="unspecified parameter error!", id=ERROR.PARAMETER):
        privacyIDEAError.__init__(self, description=description, id=id)


class DatabaseError(privacyIDEAError):
    """Error in the database layer"""

    def __init__(self, description="database error!", eid=ERROR.DATABASE):
        privacyIDEAError.__init__(self, description=description, id=eid)


class ResolverError(privacyIDEAError):
    """Error in user resolver"""
    def __init__(self, description="resolver error!", eid=ERROR.RESOLVER):
        privacyIDEAError.__init__(self, description=description, id=eid)


class ContainerError(privacyIDEAError):
    def __init__(self, description="container error!", eid=ERROR.CONTAINER):
        privacyIDEAError.__init__(self, description=description, id=eid)


class ContainerNotRegistered(ContainerError):
    def __init__(self, description="container is not registered error!", eid=ERROR.CONTAINER_NOT_REGISTERED):
        ContainerError.__init__(self, description=description, eid=eid)


class ContainerInvalidChallenge(ContainerError):
    def __init__(self, description="container challenge error!", eid=ERROR.CONTAINER_INVALID_CHALLENGE):
        ContainerError.__init__(self, description=description, eid=eid)


class ContainerRollover(ContainerError):
    def __init__(self, description="container rollover error", eid=ERROR.CONTAINER_ROLLOVER):
        ContainerError.__init__(self, description=description, eid=eid)



#  privacyIDEA
#  Jul 18, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from privacyidea.lib.applications import MachineApplicationBase
from privacyidea.lib.utils import hexlify_and_unicode
from privacyidea.lib.crypto import geturandom
from privacyidea.lib.token import get_tokens
from privacyidea.lib.policy import TYPE
import logging
log = logging.getLogger(__name__)


class MachineApplication(MachineApplicationBase):
    """
    This is the application for LUKS.

    required options:
        slot
        partition
    """
    application_name = "luks"

    @staticmethod
    def get_authentication_item(token_type,
                                serial,
                                challenge=None,
                                options=None,
                                filter_param=None,
                                user_agent=None):
        """
        :param token_type: the type of the token. At the moment
                           we only support yubikeys, tokentype "TOTP".
        :param serial:     the serial number of the token.
                           The challenge response token needs to start with
                           "UBOM".
        :param challenge:  A challenge, for which a response get calculated.
                           If none is presented, we create one.
        :type challenge:   hex string
        :return auth_item: For Yubikey token type it
                           returns a dictionary with a "challenge" and
                           a "response".
        """
        ret = {}
        options = options or {}
        if token_type.lower() == "totp" and serial.startswith("UBOM"):
            # create a challenge of 32 byte
            # Although the yubikey is capable of doing 64byte challenges
            # the hmac module calculates different responses for 64 bytes.
            if challenge is None:
                challenge = geturandom(32)
                challenge_hex = hexlify_and_unicode(challenge)
            else:
                challenge_hex = challenge
            ret["challenge"] = challenge_hex
            # create the response. We need to get
            # the HMAC key and calculate a HMAC response for
            # the challenge
            toks = get_tokens(serial=serial, active=True)
            if len(toks) == 1:
                # tokenclass is a TimeHmacTokenClass
                (_r, _p, otp, _c) = toks[0].get_otp(challenge=challenge_hex,
                                                        do_truncation=False)
                ret["response"] = otp
        else:
            log.info("Token %r, type %r is not supported by "
                     "LUKS application module" % (serial, token_type))

        return ret

    @staticmethod
    def get_options():
        """
        returns a dictionary with a list of required and optional options
        """
        options = {"totp":
                       {'slot': {'type': TYPE.INT,
                                 'value': [0, 1, 2, 3, 4, 5, 6, 7]},
                        'partition': {'type': TYPE.STRING}}
                   }
        return options



#  2015-04-22 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writup
#
# License:  AGPLv3
# (c) 2015. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This module contains the connectors to Certificate Authorities.
In this first implementation it is only a local certificate authority.

This module is tested in tests/test_lib_caconnector.py
"""

# This takes a set of all CA Connector modules.
AvailableCAConnectors = []


class BaseCAConnector(object):
    def revoke_cert(self, certificate, request_id=None, reason=None):
        """
        Revoke the specified certificate. At this point only the database
        index.txt is updated.

        :param certificate: The certificate to revoke
        :type certificate: Either takes X509 object or a PEM encoded
            certificate (string)
        :param request_id: The Id of the certificate in the certificate authority
        :type request_id: int
        :param reason: One of the available reasons the certificate gets revoked
        :type reason: basestring
        :return: Returns the serial number of the revoked certificate. Otherwise
            an error is raised.
        """
        pass

    def sign_request(self, csr, options=None):
        """
        Signs a certificate request with the key of the CA.

        options can be
        WorkingDir: The directory where the configuration like openssl.cnf
        can be found.
        CSRDir: The directory, where to save the certificate signing
        requests. This is relative to the WorkingDir.
        CertificateDir: The directory where to save the certificates. This is
        relative to the WorkingDir.

        :param csr: Certificate signing request
        :type csr: PEM string or SPKAC
        :param options: Additional options like the validity time or the
            template or spkac=1
        :type options: dict
        :return: Returns a return value and the certificate
        :rtype: (int, x509)
        """
        pass

    def create_crl(self, publish=True, check_validity=False):
        """
        Create and Publish the CRL.

        :param publish: Whether the CRL should be published at its CDPs
        :param check_validity: Only create a new CRL, if the old one is about to
            expire. Therefore the overlap period and the remaining runtime of
            the CRL is checked. If the remaining runtime is smaller than the
            overlap period, we recreate the CRL.
        :return: the CRL location or None, if no CRL was created
        """
        pass

    def request_cert(self):
        """
        create key server side
        create key on client side
        via PKCS10
        :return:
        """
        pass

    def get_cr_status(self, request_id):
        """
        If a certificate needs a CA manager approval the request is in a pending state.
        This method fetches the state of a requested certificate.
        This way we can know, if the certificate was issued in the meantime.

        :return:
        """
        pass

    def get_issued_certificate(self, request_id):
        """
        If get_csr_status returned the information that the certificate has been enrolled,
        we can fetch the issued certificate.

        :param request_id: The id of the original certificate request
        :return: The certificate as PEM string
        """
        pass

    @classmethod
    def get_caconnector_description(cls):
        """
        Return the description of this CA connectors.
        This contains the name as a key and the possible parameters.

        :return: resolver description dict
        :rtype:  dict
        """
        return {}

    def set_config(self, config=None):
        """
        Set the configuration of the
        :param config: A dict with specific config values
        :return:
        """
        pass

    def get_templates(self):
        """
        Return a dictionary of available certificate templates.
        The names of the certificate templates are the keys of the dict.

        Depending on the user we could return different templates.

        :return:
        """
        return {}

    def get_config(self, config):
        """
        This method helps to format the config values of the CA Connector.
        :param config: The configuration as it is stored in the database
        :type config: dict
        :return:
        """
        return config

    def get_specific_options(self):
        """
        Returns a dict of additional options for a specific connector instance.
        :return:
        """
        return {}



#  2019-11-07 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             initial code for writing audit information to a file
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """The Container Audit Module allows to write audit information to several different
audit modules at the same time. E.g. it can write audit information to the SQL Audit Module and to the
Logger Audit Module. This way audit information can be saved in the SQL database and at the same time
be passed to a file or external services via the Python logging facility.

The Container Audit Module is configured like this:

    PI_AUDIT_MODULE = 'privacyidea.lib.auditmodules.containeraudit'
    PI_AUDIT_CONTAINER_WRITE = ['privacyidea.lib.auditmodules.sqlaudit','privacyidea.lib.auditmodules.loggeraudit']
    PI_AUDIT_CONTAINER_READ = 'privacyidea.lib.auditmodules.sqlaudit'

You also have to provide the configuration parameters for the referenced audit modules.

"""

import logging
from privacyidea.lib.auditmodules.base import (Audit as AuditBase)
from privacyidea.lib.utils import get_module_class


log = logging.getLogger(__name__)


class Audit(AuditBase):
    """
    This is the ContainerAudit module, which writes the audit entries
    to a list of audit modules.
    """

    def __init__(self, config=None, startdate=None):
        super(Audit, self).__init__(config, startdate)
        self.name = "containeraudit"
        write_conf = self.config.get('PI_AUDIT_CONTAINER_WRITE')
        read_conf = self.config.get('PI_AUDIT_CONTAINER_READ')
        # Initialize all modules
        self.write_modules = [get_module_class(audit_module, "Audit", "log")(config, startdate)
                              for audit_module in write_conf]
        self.read_module = get_module_class(read_conf, "Audit", "log")(config, startdate)
        if not self.read_module.is_readable:
            log.warning("The specified PI_AUDIT_CONTAINER_READ {0!s} is not readable.".format(self.read_module))

    @property
    def has_data(self):
        return any([x.has_data for x in self.write_modules])

    def log(self, param: dict[str, str]):
        """
        Call the log method for all writeable modules
        """
        for module in self.write_modules:
            module.log(param)

    def add_to_log(self, param, add_with_comma=False):
        """
        Call the add_to_log method for all writeable modules
        """
        for module in self.write_modules:
            module.add_to_log(param, add_with_comma)

    def add_policy(self, policyname):
        """
        Call the add_policy method for all writeable modules
        """
        for module in self.write_modules:
            module.add_policy(policyname)

    def search(self, search_dict, page_size=15, page=1, sortorder="asc",
               timelimit=None):
        """
        Call the search method for the one readable module
        """
        return self.read_module.search(search_dict, page_size=page_size, page=page,
                                       sortorder=sortorder, timelimit=timelimit)

    def get_count(self, search_dict, timedelta=None, success=None):
        """
        Call the count method for the one readable module
        """
        return self.read_module.get_count(search_dict, timedelta=timedelta, success=success)

    def csv_generator(self, param=None, user=None, timelimit=None):
        """
        Call the csv_generator method for the one readable module
        """
        return self.read_module.csv_generator(param=param, user=user,
                                              timelimit=timelimit)

    def get_total(self, param, AND=True, display_error=True, timelimit=None):
        """
        Call the total method for the one readable module
        """
        return self.read_module.get_total(param, AND=AND, display_error=display_error, timelimit=timelimit)

    def finalize_log(self):
        """
        Call the finalize method of all writeable audit modules
        """
        for module in self.write_modules:
            module.finalize_log()



#  Mar, 12 2020 Bruno Cascio
#  http://www.privacyidea.org
#
#  product:  PrivacyIDEA
#  module:   httpresolver
#  tool:     HTTPResolver
#  edition:  Comunity Edition
#
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .UserIdResolver import UserIdResolver
import requests
import logging
import json
from urllib.parse import urlencode
from pydash import get

ENCODING = "utf-8"

__name__ = "HTTP_RESOLVER"

log = logging.getLogger(__name__)


class HTTPResolver(UserIdResolver):

    fields = {
        "endpoint": 1,
        "method": 1,
        "requestMapping": 1,
        "headers": 1,
        "responseMapping": 1,
        "hasSpecialErrorHandler": 0,
        "errorResponse": 0
    }

    def __init__(self):
        super(HTTPResolver, self).__init__()
        self.config = {}

    @staticmethod
    def getResolverClassType():
        """
        provide the resolver type for registration
        """
        return 'httpresolver'

    @staticmethod
    def getResolverType():
        """
        getResolverType - return the type of the resolver

        :return: returns the string 'ldapresolver'
        :rtype:  string
        """
        return HTTPResolver.getResolverClassType()

    @classmethod
    def getResolverClassDescriptor(cls):
        """
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        """
        descriptor = {}
        typ = cls.getResolverClassType()
        descriptor['clazz'] = "useridresolver.HTTPResolver.HTTPResolver"
        descriptor['config'] = {
            'endpoint': 'string',
            'method': 'string',
            'headers': 'string',
            'requestMapping': 'string',
            'responseMapping': 'string',
            'hasSpecialErrorHandler': 'bool',
            'errorResponse': 'string',
        }
        return {typ: descriptor}

    @staticmethod
    def getResolverDescriptor():
        """
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        """
        return HTTPResolver.getResolverClassDescriptor()

    def getUserId(self, loginName):
        """
        This method only echo's the loginName parameter
        """
        return loginName

    def getUsername(self, userid):
        """
        This method only echo's the userid parameter
        """
        return userid

    def getUserInfo(self, userid):
        """
        This function returns all user information for a given user object
        identified by UserID.
        :param userid: ID of the user in the resolver
        :type userid: int or string
        :return:  dictionary, if no object is found, the dictionary is empty
        :rtype: dict
        """
        return self._getUser(userid)

    def getUserList(self, searchDict=None):
        """
        Since it is an HTTP resolver,
        users are not stored in the database
        """
        return []

    def getResolverId(self):
        """
        get resolver specific information
        :return: the resolver identifier string - empty string if not exist
        """
        return self.config['endpoint'] if 'endpoint' in self.config else ''

    def loadConfig(self, config):
        """
        Load the configuration from the dict into the Resolver object.
        If attributes are missing, need to set default values.
        If required attributes are missing, this should raise an
        Exception.

        :param config: The configuration values of the resolver
        :type config: dict
        """
        self.config = config
        return self

    @classmethod
    def testconnection(cls, param):
        """
        This function lets you test if the parameters can be used to create a
        working resolver. Also, you can use it anytime you see if the API is
        running as expected.
        The implementation should try to make a request to the HTTP API and verify
        if user can be retrieved.
        In case of success it should return the raw http response.

        :param param: The parameters that should be saved as the resolver
        :type param: dict
        :return: returns True in case of success and a raw response
        :rtype: tuple
        """
        desc = ""
        success = False
        try:
            resolver = HTTPResolver()
            resolver.loadConfig(param)
            response = resolver._getUser(param.get('testUser'))
            desc = response
            success = True
        except Exception as e:
            success = False
            desc = "failed: {0!s}".format(e)
        return success, desc

    #
    #   Private methods
    #
    def _getUser(self, userid):
        param = self.config
        method = param.get('method').lower()
        endpoint = param.get('endpoint')
        requestMappingJSON = json.loads(param.get('requestMapping').replace("{userid}", userid))
        responseMapping = json.loads(param.get('responseMapping'))
        headers = json.loads(param.get('headers', '{}'))
        hasSpecialErrorHandler = bool(param.get('hasSpecialErrorHandler'))
        errorResponse = json.loads(param.get('errorResponse', '{}'))

        if method == "post":
            httpResponse = requests.post(endpoint, json=requestMappingJSON, headers=headers, timeout=60)
        else:
            httpResponse = requests.get(endpoint, urlencode(requestMappingJSON), headers=headers, timeout=60)

        # Raises HTTPError, if one occurred.
        httpResponse.raise_for_status()

        jsonHTTPResponse = httpResponse.json()

        if hasSpecialErrorHandler:
            # verify if error response mapping is a subset of the json http response
            if all([x in jsonHTTPResponse.items() for x in errorResponse.items()]):
                log.error(jsonHTTPResponse)
                raise Exception('Received an error while searching for user: %s' % userid)

        # Create mapped response with response mapping resolver input
        response = {}
        for pi_user_key, value in responseMapping.items():
            if value.startswith('{') and value.endswith('}'):
                response[pi_user_key] = get(jsonHTTPResponse, value[1:-1])
            else:
                response[pi_user_key] = value

        return response



"""
2015-06-04 Create REDIS mock
           Cornelius Kölbel <cornelius@privacyidea.org>

Original responses.py is:
Copyright 2013 Dropbox, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import namedtuple

from collections.abc import Sequence, Sized

from .smtpmock import get_wrapped

Call = namedtuple('Call', ['setex', 'get'])

_wrapper_template = """\
def wrapper%(signature)s:
    with redismock:
        return func%(funcargs)s
"""


class CallList(Sequence, Sized):
    def __init__(self):
        self._calls = []

    def __iter__(self):
        return iter(self._calls)

    def __len__(self):
        return len(self._calls)

    def __getitem__(self, idx):
        return self._calls[idx]

    def setdata(self, request, response):
        self._calls.append(Call(request, response))

    def reset(self):
        self._calls = []


class Redis(object):

    def __init__(self):
        self.dictionary = {}

    def get(self, value):
        return self.dictionary.get(value)

    def setex(self, key, value=None, time=None):
        self.dictionary[key] = value
        return True

    def set_data(self, data):
        self.dictionary = data


class RedisMock(object):

    def __init__(self):
        self._calls = CallList()
        self.data = {}
        self.reset()

    def reset(self):
        self._calls.reset()

    @property
    def calls(self):
        return self._calls

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.stop()
        self.reset()

    def activate(self, func):
        evaldict = {'redismock': self, 'func': func}
        return get_wrapped(func, _wrapper_template, evaldict)

    def set_data(self, data):
        self.data = data

    def start(self):
        import mock

        def unbound_on_Redis(hostname):
            self.redis_obj = Redis()
            self.redis_obj.set_data(self.data)
            return self.redis_obj

        self._patcher = mock.patch('redis.Redis',
                                   unbound_on_Redis)
        self._patcher.start()

    def stop(self):
        self._patcher.stop()

# expose default mock namespace
mock = _default_mock = RedisMock()
__all__ = []
for __attr in (a for a in dir(_default_mock) if not a.startswith('_')):
    __all__.append(__attr)
    globals()[__attr] = getattr(_default_mock, __attr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# 2018-06-29 Friedrich Weber <friedrich.weber@netknights.it>
#            Implement periodic task runner
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """
This script is meant to be invoked periodically by the system cron daemon.
It runs periodic tasks that are specified in the database.
"""

import click
from datetime import datetime
from dateutil import tz
from flask import current_app
import json
import sys
import traceback
import warnings

from privacyidea.cli import create_silent_app, NoPluginsFlaskGroup
from privacyidea.lib.config import get_privacyidea_node
from privacyidea.lib.periodictask import (get_scheduled_periodic_tasks,
                                          execute_task, get_periodic_tasks,
                                          get_periodic_task_by_name,
                                          set_periodic_task_last_run)
from privacyidea.lib.utils import get_version_number

warnings.simplefilter("ignore")

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def print_stdout(*args, **kwargs):
    """
    Print to stdout, except if "cron mode" has been activated.
    """
    if not current_app.config.get("cron_mode", False):
        click.echo(*args, **kwargs)


def print_stderr(*args, **kwargs):
    """
    Print to stderr.
    """
    click.echo(*args, file=sys.stderr, **kwargs)


def get_node_name(node):
    """
    Determine the node name. If no node name is given, read it from the app config.

    :param node: node name given by the user (can be None)
    :return:
    """
    if node is not None:
        return node
    else:
        return get_privacyidea_node()


def run_task_on_node(ptask, node):
    """
    Run a periodic task (given as a dictionary) on the given node.
    In case of success, write the last successful run to the database. Catch any exceptions.

    :param ptask: task as a dictionary
    :param node: Node name
    """
    try:
        print_stdout("Running {!r} ...".format(ptask["name"]), nl=False)
        result = execute_task(ptask["taskmodule"], ptask["options"])
    except Exception as e:
        print_stderr('Caught exception when running {!r}: {!r}'.format(ptask["name"], e))
        print_stderr(f"{traceback.format_exc()}")
        result = False
    if result:
        current_time = datetime.now(tz.tzlocal())
        print_stdout('Task {!r} on node {!r} exited successfully. Noting this '
                     'in the database ...'.format(ptask["name"], node))
        set_periodic_task_last_run(ptask["id"], node, current_time)
    else:
        print_stderr('Task {!r} on node {!r} did not run '
                     'successfully.'.format(ptask["name"], node))
        print_stderr('This unsuccessful run is not recorded in the database.')
        if not ptask.get("retry_if_failed"):
            current_time = datetime.now(tz.tzlocal())
            set_periodic_task_last_run(ptask["id"], node, current_time)
    return result


@click.group(cls=NoPluginsFlaskGroup, create_app=create_silent_app,
             context_settings=CONTEXT_SETTINGS, add_default_commands=False,
             epilog='Check out our docs at https://privacyidea.readthedocs.io/ for more details')
def cli():
    """
\b
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \\/ __/ _ |
  / _ \\/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\\_,_/\\__/\\_, /___/____/___/_/ |_|  Cron
/_/                       /___/

This script is meant to be invoked periodically by the system cron daemon.
It runs periodic tasks that are specified in the database.
"""
    pass


@cli.command("run_manually")
@click.option("-n", "--node", "node_string",
              help="Override the node name (read from privacyIDEA config by default)")
@click.option("-t", "--task", "task_name",
              help="Run the specified task",
              required=True)
def run_manually(node_string, task_name):
    """
    Manually run a periodic task.
    BEWARE: This does not check whether the task is active, or whether it should
    run on the given node at all.
    """
    node = get_node_name(node_string)
    ptask = get_periodic_task_by_name(task_name)
    run_task_on_node(ptask, node)


@cli.command("list")
def list_tasks():
    """
    Show a list of available tasks that could be run.
    """
    line_format = ('{active!s:7.7} {id:3} {name:16.16}\t{interval:16.16}\t{taskmodule:16}'
                   '\t{node_list:20}\t{options_json}')
    heading = line_format.format(
        active="Active",
        id="ID",
        name="Name",
        interval="Interval",
        taskmodule="Task Module",
        node_list="Nodes",
        options_json="Options",
    )
    print_stdout(heading)
    print_stdout("=" * 120)
    for ptask in get_periodic_tasks():
        print_stdout(line_format.format(node_list=', '.join(ptask["nodes"]),
                                        options_json=json.dumps(ptask["options"]),
                                        **ptask))


@cli.command("run_scheduled")
@click.option("-d", "--dryrun",
              is_flag=True,
              help="Do not run any tasks, only show what would be done")
@click.option("-n", "--node", "node_string",
              help="Override the node name (read from privacyIDEA config by default)")
@click.option("-c", "--cron", "cron_mode", is_flag=True,
              help="Run in 'cron mode', i.e. do not write to stdout, but write errors to stderr")
def run_scheduled(node_string=None, dryrun=False, cron_mode=False):
    """
    Execute all periodic tasks that are scheduled to run.
    """
    current_app.config['cron_mode'] = cron_mode
    node = get_node_name(node_string)
    current_time = datetime.now(tz.tzlocal())
    scheduled_tasks = get_scheduled_periodic_tasks(node, current_time)
    if scheduled_tasks:
        print_stdout("The following tasks are scheduled to run on node {!s}:".format(node))
        print_stdout()
        for ptask in scheduled_tasks:
            print_stdout("  {name} ({interval!r}, {taskmodule})".format(**ptask))

        print_stdout()
        if not dryrun:
            results = []
            for ptask in scheduled_tasks:
                result = run_task_on_node(ptask, node)
                results.append(result)
            if all(results):
                print_stdout("All scheduled tasks executed successfully.")
            else:
                print_stderr("Some tasks exited with errors.")
                sys.exit(1)
        else:
            print_stdout("Not running any tasks because --dryrun was passed.")
    else:
        print_stdout("There are no tasks scheduled on node {!s}.".format(node))


if __name__ == '__main__':
    cli()



import os
import logging
import secrets
import string
log = logging.getLogger(__name__)
basedir = os.path.abspath(os.path.dirname(__file__))
basedir = "/".join(basedir.split("/")[:-1]) + "/"


pubtest_key = b"""-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0stF+PVh/qR7I82ywgDW
X2rPxKnlBugU33SAbxZjfNz+aKYYM+irKJ0IdTOLtQ85DzZaEhocqO7cg1Qw85Ac
gU8NEIUdRQdHilO558Mc/E95jDeiI7AbzuLlG2gFyMHkObqwoA2N9ARk9vck+pIo
gCVhZfHs+ifRqdOMzbGNtWp8XGwpp9YLEFTZ2ZBloWJ6kkijatXwY5nTlP5ugbEU
cKMamksTyvRdZFryKwK3HfAQoL6fJAfB3u4NVp5nfcuD180MKcGk8CR9zwuXkRX5
m+qnl1W6iD0z5ws/SrtWBJiMuFm3hGaXdQLZVABSqgmvaru3GZbldlQU/0pcV855
NwIDAQAB
-----END PUBLIC KEY-----"""

non_matchin_pubkey = b"""-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0R9Nos93gt6JdiY/Q+ac
WuukYcwr+JlCDa+RX8UDuKfF3wVkSjwAl6s3ADrEnRIgi9qCQD52o29YEVugftVh
HKiC+Gw5Cf95hkWKLl1MFoI6JUMDnNtPlPJrsCwLCi79TV+24BZ0v19P9ivxDa+c
bHUR+vUHbvYfhzpf5Zq8gHTNEQda6+ufcC48VjCdJwOiJJcjNZMZhYrQBoYyCHxm
HUb6rdpz/AsBSMV3+kZiDqv/NA5dgR8dXAK2or8JSK/ghw8c3qgt+yLW+1g9FDao
6Bn6LrQkUkMv/cANYsiVk/pnxtsb11cEbYNyv2VQd/lbCCMwK/rkuipZxu8XJl6g
WQIDAQAB
-----END PUBLIC KEY-----
"""


def _random_password(size):
    log.info("SECRET_KEY not set in config. Generating a random key.")
    passwd = [secrets.choice(string.ascii_lowercase + \
                             string.ascii_uppercase + string.digits) for _x in range(size)]
    return "".join(passwd)


class Config(object):
    SECRET_KEY = os.environ.get('SECRET_KEY')
    # SQL_ALCHEMY_DATABASE_URI = "mysql://privacyidea:XmbSrlqy5d4IS08zjz"
    # "GG5HTt40Cpf5@localhost/privacyidea"
    PI_ENCFILE = os.path.join(basedir, "tests/testdata/enckey")
    PI_HSM = "default"
    PI_AUDIT_MODULE = "privacyidea.lib.auditmodules.sqlaudit"
    PI_AUDIT_KEY_PRIVATE = os.path.join(basedir, "tests/testdata/private.pem")
    PI_AUDIT_KEY_PUBLIC = os.path.join(basedir, "tests/testdata/public.pem")
    PI_LOGFILE = "privacyidea.log"
    PI_LOGLEVEL = logging.INFO
    CACHE_TYPE = "simple"
    PI_EXTERNAL_LINKS = True
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    # PI_GNUPG_HOME = "gpg"
    # PI_LOGO = "otherlogo.png"
    # PI_AUDIT_SQL_URI = sqlite://
    PI_VASCO_LIBRARY = None
    PI_CUSTOM_CSS = False


class DevelopmentConfig(Config):
    DEBUG = True
    SECRET_KEY = os.environ.get('SECRET_KEY') or 't0p s3cr3t'
    SQLALCHEMY_DATABASE_URI = os.environ.get('DEV_DATABASE_URL') or \
        'sqlite:///' + os.path.join(basedir, 'data-dev.sqlite')
    PI_LOGLEVEL = logging.DEBUG
    PI_TRANSLATION_WARNING = "[Missing]"


class TestingConfig(Config):
    TESTING = True
    # This is used to encrypt the auth token
    SUPERUSER_REALM = ['adminrealm']
    SECRET_KEY = 'secret'  # nosec B105 # used for testing
    SQLALCHEMY_DATABASE_URI = os.environ.get('TEST_DATABASE_URL') or \
        'sqlite:///' + os.path.join(basedir, 'data-test.sqlite')
    # This is used to encrypt the admin passwords
    PI_PEPPER = ""
    # This is only for testing encrypted files
    PI_ENCFILE_ENC = "tests/testdata/enckey.enc"
    PI_LOGLEVEL = logging.INFO
    PI_GNUPG_HOME = "tests/testdata/gpg"
    PI_AUDIT_SQL_TRUNCATE = True
    CACHE_TYPE = "None"
    PI_SCRIPT_HANDLER_DIRECTORY = "tests/testdata/scripts/"
    PI_NOTIFICATION_HANDLER_SPOOLDIRECTORY = "tests/testdata/"
    PI_NODE_UUID = "8e4272a9-9037-40df-8aa3-976e4a04b5a9"
    PI_NODE = "Node1"
    PI_ENGINE_REGISTRY_CLASS = "null"
    PI_EMAIL_VALIDATOR_MODULES = ["tests.testdata.gmailvalidator"]
    PI_TRUSTED_JWT = [{"public_key": pubtest_key,
                       "algorithm": "HS256",
                       "role": "user",
                       "realm": "realm1",
                       "username": "userA",
                       "resolver": "resolverX"},
                      {"public_key": non_matchin_pubkey,
                       "algorithm": "RS256",
                       "role": "user",
                       "realm": "realm1",
                       "username": "userA",
                       "resolver": "resolverX"},
                      {"public_key": pubtest_key,
                       "algorithm": "RS256",
                       "role": "user",
                       "realm": "realmX",
                       "resolver": "resolverX",
                       "username": "h.*s"},
                      {"public_key": pubtest_key,
                       "algorithm": "RS256",
                       "role": "user",
                       "realm": "realm1",
                       "username": "userA",
                       "resolver": "resolverX"}]


class AltUIConfig(TestingConfig):
    PI_INDEX_HTML = "testui.html"
    PI_STATIC_FOLDER = "../tests/testdata/altstatic"
    PI_TEMPLATE_FOLDER = "../tests/testdata/altstatic/templates"


class ProductionConfig(Config):
    SQLALCHEMY_DATABASE_URI = os.environ.get('DATABASE_URL') or \
        'sqlite:///' + os.path.join(basedir, 'data.sqlite')
    #SQLALCHEMY_DATABASE_URI = "mysql://pi2:pi2@localhost/pi2"
    # This is used to encrypt the auth_token
    SECRET_KEY = os.environ.get('SECRET_KEY') or _random_password(24)
    # This is used to encrypt the admin passwords
    PI_PEPPER = "Never know..."
    # This is used to encrypt the token data and token passwords
    PI_ENCFILE = os.path.join(basedir, "enckey")
    # This is used to sign the audit log
    PI_AUDIT_KEY_PRIVATE = os.path.join(basedir, "private.pem")
    PI_AUDIT_KEY_PUBLIC = os.path.join(basedir, "public.pem")
    PI_LOGLEVEL = logging.INFO
    SUPERUSER_REALM = ['superuser']


class HerokuConfig(Config):
    SQLALCHEMY_DATABASE_URI = "postgres://mvfkmtkwzuwojj:" \
                              "wqy_btZE3CPPNWsmkfdmeorxy6@" \
                              "ec2-54-83-0-61.compute-1." \
                              "amazonaws.com:5432/d6fjidokoeilp6"
    #SQLALCHEMY_DATABASE_URI = "mysql://pi2:pi2@localhost/pi2"
    # This is used to encrypt the auth_token
    SECRET_KEY = os.environ.get('SECRET_KEY') or 't0p s3cr3t'
    # This is used to encrypt the admin passwords
    PI_PEPPER = "Never know..."
    # This is used to encrypt the token data and token passwords
    PI_ENCFILE = os.path.join(basedir, "deploy/heroku/enckey")
    # This is used to sign the audit log
    PI_AUDIT_KEY_PRIVATE = os.path.join(basedir,
                                        "deploy/heroku/private.pem")
    PI_AUDIT_KEY_PUBLIC = os.path.join(basedir, "deploy/heroku/public.pem")
    SUPERUSER_REALM = ['superuser']


config = {
    'development': DevelopmentConfig,
    'testing': TestingConfig,
    'production': ProductionConfig,
    'default': DevelopmentConfig,
    'heroku': HerokuConfig,
    'altUI': AltUIConfig
}



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
import json
import logging

from flask import Blueprint, request, g

from privacyidea.api.auth import admin_required
from privacyidea.api.lib.prepolicy import (check_base_action, prepolicy, check_user_params, check_token_action,
                                           check_admin_tokenlist, check_container_action, check_token_list_action,
                                           check_container_register_rollover, container_registration_config,
                                           smartphone_config, check_client_container_action, hide_tokeninfo,
                                           check_client_container_disabled_action)
from privacyidea.api.lib.utils import send_result, getParam, required
from privacyidea.lib.container import (find_container_by_serial, init_container, get_container_classes_descriptions,
                                       get_container_token_types, get_all_containers, add_container_info,
                                       set_container_description, set_container_states, set_container_realms,
                                       delete_container_by_serial, assign_user, unassign_user, add_token_to_container,
                                       add_multiple_tokens_to_container, remove_token_from_container,
                                       remove_multiple_tokens_from_container,
                                       create_container_dict, create_endpoint_url,
                                       create_container_template,
                                       get_templates_by_query, create_container_tokens_from_template, get_template_obj,
                                       set_default_template, get_container_template_classes,
                                       create_container_from_db_object,
                                       compare_template_with_container, unregister,
                                       finalize_registration, init_container_rollover,
                                       get_container_realms,
                                       add_not_authorized_tokens_result, get_offline_token_serials,
                                       delete_container_info)
from privacyidea.lib.containerclass import TokenContainerClass
from privacyidea.lib.containers.container_info import TokenContainerInfoData, PI_INTERNAL
from privacyidea.lib.error import ParameterError, ContainerNotRegistered, ContainerError
from privacyidea.lib.event import event
from privacyidea.lib.log import log_with
from privacyidea.lib.policy import ACTION
from privacyidea.lib.token import regenerate_enroll_url
from privacyidea.lib.user import get_user_from_param

container_blueprint = Blueprint('container_blueprint', __name__)
log = logging.getLogger(__name__)

__doc__ = """
API for managing token containers
"""


@container_blueprint.route('/', methods=['GET'])
@prepolicy(check_base_action, request, action=ACTION.CONTAINER_LIST)
@prepolicy(check_admin_tokenlist, request, ACTION.CONTAINER_LIST)
@prepolicy(check_admin_tokenlist, request, ACTION.TOKENLIST)
@prepolicy(hide_tokeninfo, request)
@log_with(log)
def list_containers():
    """
    Get containers depending on the query parameters. If pagesize and page are not provided, all containers are returned
    at once.

    :query user: Username of a user assigned to the containers
    :query container_serial: Serial of a single container
    :query type: Type of the containers to return
    :query token_serial: Serial of a token assigned to the container
    :query template: Name of the template the container is created from
    :query sortby: Sort by a container attribute (serial or type)
    :query sortdir: Sort direction (asc or desc)
    :query pagesize: Number of containers per page
    :query page: Page number
    :jsonparam no_token: no_token=1: Do not return tokens assigned to the container
    """
    param = request.all_data
    user = request.User
    cserial = getParam(param, "container_serial", optional=True)
    ctype = getParam(param, "type", optional=True)
    token_serial = getParam(param, "token_serial", optional=True)
    sortby = getParam(param, "sortby", optional=True, default="serial")
    sortdir = getParam(param, "sortdir", optional=True, default="asc")
    psize = int(getParam(param, "pagesize", optional=True) or 0)
    page = int(getParam(param, "page", optional=True) or 0)
    no_token = getParam(param, "no_token", optional=True, default=False)
    template = getParam(param, "template", optional=True)
    logged_in_user_role = g.logged_in_user.get("role")
    allowed_container_realms = getattr(request, "pi_allowed_container_realms", None)
    allowed_token_realms = getattr(request, "pi_allowed_realms", None)
    hide_token_info = getParam(param, "hidden_tokeninfo", optional=True)

    result = get_all_containers(user=user, serial=cserial, ctype=ctype, token_serial=token_serial,
                                realms=allowed_container_realms, template=template,
                                sortby=sortby, sortdir=sortdir,
                                pagesize=psize, page=page)

    containers = create_container_dict(result["containers"], no_token, user, logged_in_user_role, allowed_token_realms,
                                       hide_token_info=hide_token_info)
    result["containers"] = containers

    g.audit_object.log({"success": True,
                        "info": f"allowed_container_realms={allowed_container_realms}, "
                                f"allowed_token_realms={allowed_token_realms}"})
    return send_result(result)


@container_blueprint.route('<string:container_serial>/assign', methods=['POST'])
@prepolicy(check_user_params, request, action=ACTION.CONTAINER_ASSIGN_USER)
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_ASSIGN_USER)
@event('container_assign', request, g)
@log_with(log)
def assign(container_serial):
    """
    Assign a container to a user.

    :param container_serial: serial of the container
    :jsonparam user: Username of the user
    :jsonparam realm: Name of the realm of the user
    """
    user = get_user_from_param(request.all_data, required)
    res = assign_user(container_serial, user)

    container = find_container_by_serial(container_serial)
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "success": res}
    g.audit_object.log(audit_log_data)
    return send_result(res)


@container_blueprint.route('<string:container_serial>/unassign', methods=['POST'])
@prepolicy(check_user_params, request, action=ACTION.CONTAINER_UNASSIGN_USER)
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_UNASSIGN_USER)
@event('container_unassign', request, g)
@log_with(log)
def unassign(container_serial):
    """
    Unassign a user from a container

    :param container_serial: serial of the container
    :jsonparam user: Username of the user
    :jsonparam realm: Realm of the user
    """
    user = get_user_from_param(request.all_data, required)
    res = unassign_user(container_serial, user)

    container = find_container_by_serial(container_serial)
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "success": res}
    g.audit_object.log(audit_log_data)
    return send_result(res)


@container_blueprint.route('init', methods=['POST'])
@prepolicy(check_admin_tokenlist, request, action=ACTION.CONTAINER_CREATE)
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_CREATE)
@event('container_init', request, g)
@log_with(log)
def init():
    """
    Create a new container.
    Raises an EnrollmentError if an invalid type or an already existing serial is provided.

    To create a container from a template, the template can either be passed as a dictionary, containing all necessary
    information for all tokens or only the name of an already existing template in the db. If both are given, a
    ParameterError is raised.

    :jsonparam description: Description for the container
    :jsonparam type: Type of the container. If the type is unknown, an error will be returned
    :jsonparam container_serial: Optional unique serial (not case-sensitive)
    :jsonparam user: Optional username to assign the container to. Requires realm param to be present as well.
    :jsonparam realm: Optional realm to assign the container to. Requires user param to be present as well.
    :jsonparam template: The template to create the container from (dictionary), optional
    :jsonparam template_name: The name of the template to create the container from, optional
    :jsonparam options: Options for the container if no template is used (dictionary), optional
    """
    user_role = g.logged_in_user.get("role")
    allowed_realms = getattr(request, "pi_allowed_realms", None)
    if user_role == "admin" and allowed_realms:
        req_realm = getParam(request.all_data, "realm", optional=True)
        if not req_realm or req_realm == "":
            # The container has to be in one realm the admin is allowed to manage
            request.all_data["realm"] = allowed_realms[0]

    init_res = init_container(request.all_data)
    serial = init_res["container_serial"]
    res = {"container_serial": serial}
    container = find_container_by_serial(serial)

    # Template handling
    template_tokens = init_res["template_tokens"]
    if template_tokens:
        res["tokens"] = create_container_tokens_from_template(serial, template_tokens, request, user_role)

    # Audit log
    owners = container.get_users()
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": serial,
                      "container_type": request.all_data.get("type") or "",
                      "success": True}
    if request.all_data.get("description"):
        audit_log_data["action_detail"] = f"description={request.all_data.get('description')}"
    g.audit_object.log(audit_log_data)
    return send_result(res)


@container_blueprint.route('<string:container_serial>', methods=['DELETE'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_DELETE)
@event('container_delete', request, g)
@log_with(log)
def delete(container_serial):
    """
    Delete a container.

    :param container_serial: serial of the container
    """
    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type}
    g.audit_object.log(audit_log_data)

    # Delete container
    res = delete_container_by_serial(container_serial)
    g.audit_object.log({"success": res > 0})

    return send_result(True)


@container_blueprint.route('<string:container_serial>/add', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_ADD_TOKEN)
@prepolicy(check_token_action, request, action=ACTION.CONTAINER_ADD_TOKEN)
@event('container_add_token', request, g)
@log_with(log)
def add_token(container_serial):
    """
    Add a single token to a container.

    :param container_serial: serial of the container
    :jsonparam serial: Serial of the token to add.
    """
    token_serial = getParam(request.all_data, "serial", optional=False, allow_empty=False)

    success = add_token_to_container(container_serial, token_serial)

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "serial": token_serial,
                      "success": success}
    g.audit_object.log(audit_log_data)
    return send_result(success)


@container_blueprint.route('<string:container_serial>/addall', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_ADD_TOKEN)
@prepolicy(check_token_list_action, request, action=ACTION.CONTAINER_ADD_TOKEN)
@event('container_add_token', request, g)
@log_with(log)
def add_all_tokens(container_serial):
    """
    Add multiple tokens to a container.

    :param container_serial: serial of the container
    :jsonparam serial: Comma separated list of token serials
    """
    serial = getParam(request.all_data, "serial", optional=False, allow_empty=False)
    token_serials = serial.replace(' ', '').split(',')

    res = add_multiple_tokens_to_container(container_serial, token_serials)

    not_authorized_serials = getParam(request.all_data, "not_authorized_serials", optional=True)
    res = add_not_authorized_tokens_result(res, not_authorized_serials)

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    success = False not in res.values()
    result_str = ", ".join([f"{k}: {v}" for k, v in res.items()])
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "serial": serial,
                      "success": success,
                      "info": result_str}
    g.audit_object.log(audit_log_data)
    return send_result(res)


@container_blueprint.route('<string:container_serial>/remove', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_REMOVE_TOKEN)
@prepolicy(check_token_action, request, action=ACTION.CONTAINER_REMOVE_TOKEN)
@event('container_remove_token', request, g)
@log_with(log)
def remove_token(container_serial):
    """
    Remove a single token from a container.

    :param container_serial: serial of the container
    :jsonparam serial: Serial of the token to remove.
    """
    token_serial = getParam(request.all_data, "serial", optional=False, allow_empty=False)

    success = remove_token_from_container(container_serial, token_serial)

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "serial": token_serial,
                      "success": success}
    g.audit_object.log(audit_log_data)
    return send_result(success)


@container_blueprint.route('<string:container_serial>/removeall', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_REMOVE_TOKEN)
@prepolicy(check_token_list_action, request, action=ACTION.CONTAINER_REMOVE_TOKEN)
@event('container_remove_token', request, g)
@log_with(log)
def remove_all_tokens(container_serial):
    """
    Remove multiple tokens from a container.

    :param container_serial: serial of the container
    :jsonparam serial: Comma separated list of token serials.
    """
    # allow serial to be empty if the pre-policy removes all tokens
    serial = getParam(request.all_data, "serial", optional=False, allow_empty=True)
    token_serials = serial.replace(' ', '').split(',')

    res = remove_multiple_tokens_from_container(container_serial, token_serials)

    not_authorized_serials = getParam(request.all_data, "not_authorized_serials", optional=True)
    res = add_not_authorized_tokens_result(res, not_authorized_serials)

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    success = False not in res.values()
    result_str = ", ".join([f"{k}: {v}" for k, v in res.items()])
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "serial": serial,
                      "success": success,
                      "info": result_str}
    g.audit_object.log(audit_log_data)
    return send_result(res)


@container_blueprint.route('types', methods=['GET'])
@container_blueprint.route('tokentypes', methods=['GET'])
@log_with(log)
def get_types():
    """
    Returns a dictionary with the container types as keys and their descriptions and supported token types as values.

    ::

        {
            type: { description: "Description", token_types: ["hotp", "totp", "push", "daypassword", "sms"] },
            type: { description: "Description", token_types: ["hotp", "totp", "push", "daypassword", "sms"] }
        }
    """
    descriptions = get_container_classes_descriptions()
    ttypes = get_container_token_types()
    res = {ctype: {"description": desc, "token_types": ttypes.get(ctype, [])} for ctype, desc in descriptions.items()}
    g.audit_object.log({"success": True})
    return send_result(res)


@container_blueprint.route('<string:container_serial>/description', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_DESCRIPTION)
@event('container_set_description', request, g)
@log_with(log)
def set_description(container_serial):
    """
    Set the description of a container.

    :param container_serial: Serial of the container
    :jsonparam description: New description to be set
    """
    new_description = getParam(request.all_data, "description", optional=required)
    set_container_description(container_serial, new_description)

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "action_detail": f"description={new_description}",
                      "success": True}
    g.audit_object.log(audit_log_data)
    return send_result(True)


@container_blueprint.route('<string:container_serial>/states', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_STATE)
@event('container_set_states', request, g)
@log_with(log)
def set_states(container_serial):
    """
    Set the states of a container.

    :jsonparam states: string of comma separated states
    """
    states_string = getParam(request.all_data, "states", required, allow_empty=False)
    states_string = states_string.replace(" ", "")
    states = states_string.split(",")
    res = set_container_states(container_serial, states)

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    success = False not in res.values()
    map_to_bool = ["False", "True"]
    result_str = ", ".join([f"{k}: {map_to_bool[v]}" for k, v in res.items()])
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "action_detail": f"states={states}",
                      "success": success,
                      "info": result_str}
    g.audit_object.log(audit_log_data)
    return send_result(res)


@container_blueprint.route('statetypes', methods=['GET'])
@log_with(log)
def get_state_types():
    """
    Get the supported state types as dictionary.
    The types are the keys and the value is a list containing all states that are excluded when the key state is
    selected.
    """
    state_types_exclusions = TokenContainerClass.get_state_types()
    g.audit_object.log({"success": True})
    return send_result(state_types_exclusions)


@container_blueprint.route('<string:container_serial>/realms', methods=['POST'])
@admin_required
@prepolicy(check_admin_tokenlist, request, action=ACTION.CONTAINER_REALMS)
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_REALMS)
@event('container_set_realms', request, g)
@log_with(log)
def set_realms(container_serial):
    """
    Set the realms of a container. Old realms will be deleted.

    :param container_serial: Serial of the container
    :jsonparam realms: comma separated string of realms, e.g. "realm1,realm2"
    """
    # Get parameters
    container_realms = getParam(request.all_data, "realms", required, allow_empty=True)
    realm_list = [r.strip() for r in container_realms.split(",")]
    allowed_realms = getattr(request, "pi_allowed_realms", None)

    # Set realms
    result = set_container_realms(container_serial, realm_list, allowed_realms)
    success = False not in result.values()

    # Audit log
    container = find_container_by_serial(container_serial)
    owners = container.get_users()
    if len(owners) == 1:
        g.audit_object.log({"user": owners[0].login,
                            "realm": owners[0].realm,
                            "resolver": owners[0].resolver})
    audit_log_data = {"container_serial": container_serial,
                      "container_type": container.type,
                      "action_detail": f"realms={container_realms}",
                      "success": success}
    if not success:
        result_str = ", ".join([f"{k}: {v}" for k, v in result.items() if not k == "deleted"])
        audit_log_data["info"] = f"success = {result_str}"
    g.audit_object.log(audit_log_data)
    return send_result(result)


@container_blueprint.route('<string:container_serial>/info/<key>', methods=['POST'])
@admin_required
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_INFO)
@log_with(log)
def set_container_info(container_serial, key):
    """
    Set the value of a container info key. Overwrites the old value if the key already exists.
    However, existing entries of type PI_INTERNAL are not overwritten and a PolicyError is raised.

    :param container_serial: Serial of the container
    :param key: Key of the container info
    :jsonparam value: Value to set
    """
    value = getParam(request.all_data, "value", required)
    res = add_container_info(container_serial, key, value)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "key": key,
                        "value": value,
                        "success": res})
    return send_result(res)


@container_blueprint.route('<string:container_serial>/info/delete/<key>', methods=['DELETE'])
@admin_required
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_INFO)
@log_with(log)
def delete_container_info_entry(container_serial, key):
    """
    Deletes the container info for the given key. Entries of type PI_INTERNAL can not be deleted.

    :param container_serial: Serial of the container
    :param key: Key of the container info
    """
    res = delete_container_info(container_serial, key)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "key": key,
                        "success": res[key]})
    return send_result(res[key])



@container_blueprint.route('register/initialize', methods=['POST'])
@prepolicy(check_container_register_rollover, request)
@prepolicy(container_registration_config, request)
@event('container_register_initialize', request, g)
@log_with(log)
def registration_init():
    """
    Prepares the registration of a container. It returns all information required for the container to register.

    :jsonparam container_serial: Serial of the container
    :return: Result of the registration process as dictionary. The information may differ depending on the container
        type.

    An example response for smartphones looks like this:
        ::

            {
                "container_url": {"description": "URL for privacyIDEA Container Registration",
                                  "img": <QR code>,
                                  "value": "pia://container/SMPH0006D5BC?issuer=privacyIDEA&ttl=10..."},
                "nonce": "c238392af49250804c25bbd7d86408839e91fe97",
                "time_stamp": "2024-12-20T09:53:40.158319+00:00",
                "server_url": "https://pi.net",
                "ttl": 10,
                "ssl_verify": "True",
                "key_algorithm": "secp384r1",
                "hash_algorithm": "sha256"
            }
    """
    params = request.all_data
    container_serial = getParam(params, "container_serial", required)
    container_rollover = getParam(params, "rollover", optional=True)
    container = find_container_by_serial(container_serial)
    # Params set by pre-policies
    server_url = getParam(params, "server_url")
    challenge_ttl = getParam(params, "challenge_ttl")
    registration_ttl = getParam(params, "registration_ttl")
    ssl_verify = getParam(params, "ssl_verify")

    # Audit log
    info_str = (f"server_url={server_url}, challenge_ttl={challenge_ttl}min, registration_ttl={registration_ttl}min, "
                f"ssl_verify={ssl_verify}")
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "action_detail": f"rollover={container_rollover}",
                        "info": info_str})

    # Check registration state: registration init is only allowed for None (not yet registered) and "client_wait"
    # otherwise do a rollover
    registration_state = container.get_container_info_dict().get("registration_state")
    if container_rollover:
        if registration_state not in ["registered", "rollover", "rollover_completed"]:
            raise ContainerNotRegistered("Container is not registered.")
    elif registration_state and registration_state != "client_wait":
        raise ContainerError("Container is already registered.")

    # Reset last synchronization and authentication time stamps from possible previous registration
    container.reset_last_synchronization()
    container.reset_last_authentication()

    # registration
    scope = create_endpoint_url(server_url, "container/register/finalize")
    res = container.init_registration(server_url, scope, registration_ttl, ssl_verify, params)

    if container_rollover:
        # Set registration state
        info = [TokenContainerInfoData(key="registration_state", value="rollover", info_type=PI_INTERNAL),
                TokenContainerInfoData(key="rollover_server_url", value=server_url, info_type=PI_INTERNAL),
                TokenContainerInfoData(key="rollover_challenge_ttl", value=challenge_ttl, info_type=PI_INTERNAL)]
    else:
        # save policy values in container info
        info = [TokenContainerInfoData(key="server_url", value=server_url, info_type=PI_INTERNAL),
                TokenContainerInfoData(key="challenge_ttl", value=challenge_ttl, info_type=PI_INTERNAL)]
    container.update_container_info(info)

    # Check for offline tokens
    res["offline_tokens"] = get_offline_token_serials(container)

    # Audit log
    g.audit_object.log({"success": True})

    return send_result(res)


@container_blueprint.route('register/finalize', methods=['POST'])
@event('container_register_finalize', request, g)
@prepolicy(smartphone_config, request)
@log_with(log)
def registration_finalize():
    """
    This endpoint is called from a container as second step for the registration process.
    At least the container serial has to be passed in the parameters. Further parameters might be required depending on
    the container type.

    :jsonparam container_serial: Serial of the container
    :return: Result of the registration process as dictionary. The information may differ depending on the container
        type.
    """
    params = request.all_data
    container_serial = getParam(params, "container_serial", required)

    res = finalize_registration(container_serial, params)

    # Add policy information to the response
    res["policies"] = request.all_data.get("client_policies", {})

    # Audit log
    container = find_container_by_serial(container_serial)
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "success": res})
    return send_result(res)


@container_blueprint.route('register/<string:container_serial>/terminate', methods=['POST'])
@prepolicy(check_container_action, request, action=ACTION.CONTAINER_UNREGISTER)
@event('container_unregister', request, g)
@log_with(log)
def registration_terminate(container_serial: str):
    """
    Terminates the synchronization of a container with privacyIDEA.

    :param container_serial: Serial of the container
    :return: dictionary with success information such as
        ::

            {"success": True}
    """
    container = find_container_by_serial(container_serial)

    res = unregister(container)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "success": res})

    return send_result({"success": res})


@container_blueprint.route('register/terminate/client', methods=['POST'])
@prepolicy(check_client_container_disabled_action, request, action=ACTION.DISABLE_CLIENT_CONTAINER_UNREGISTER)
@event('container_unregister', request, g)
@log_with(log)
def registration_terminate_client():
    """
    Terminates the synchronization of a container with privacyIDEA.
    This endpoint can only be called from clients that are registered at the container, providing a valid signature.

    :jsonparam container_serial: Serial of the container
    :jsonparam signature: Signature of the client
    :return: dictionary with success information such as
        ::

            {"success": True}
    """
    params = request.all_data
    container_serial = getParam(params, "container_serial", required)
    container = find_container_by_serial(container_serial)

    server_url = container.get_container_info_dict().get("server_url")
    if server_url is None:
        log.debug("Server url is not set in the container info. Container might not be registered correctly.")
        server_url = " "
    scope = create_endpoint_url(server_url, "container/register/terminate/client")
    params.update({'scope': scope})
    container.check_challenge_response(params)

    res = unregister(container)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "success": res})

    return send_result({"success": res})


@container_blueprint.route('/challenge', methods=['POST'])
@event('container_create_challenge', request, g)
@log_with(log)
def create_challenge():
    """
    Creates a challenge for a container.

    :jsonparam container_serial: Serial of the container
    :jsonparam scope: Scope of the challenge, e.g. 'https://pi.com/container/synchronize'
    :return: dictionary with the challenge information

    An example response looks like this:

    ::

        {
            "server_url": "https://pi.net"
            "nonce": "123456",
            "time_stamp": "2024-10-23T05:45:02.484954+00:00",
        }

    """
    # Get params
    params = request.all_data
    scope = getParam(params, "scope", optional=False)
    container_serial = getParam(params, "container_serial", optional=False)
    container = find_container_by_serial(container_serial)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "action_detail": f"scope={scope}"})

    container_info = container.get_container_info_dict()
    registration_state = container_info.get("registration_state")
    if registration_state not in ["registered", "rollover", "rollover_completed"]:
        raise ContainerNotRegistered(f"Container is not registered.")

    # validity time for the challenge in minutes
    challenge_ttl = int(container_info.get("challenge_ttl", "2"))

    # Create challenge
    res = container.create_challenge(scope, challenge_ttl)

    # Audit log
    g.audit_object.log({"success": True})

    return send_result(res)


@container_blueprint.route('/synchronize', methods=['POST'])
@prepolicy(smartphone_config, request)
@event('container_synchronize', request, g)
def synchronize():
    """
    Compares the client tokens with the server tokens and returns the differences. Returns a dictionary with the
    container properties and the tokens to be added or updated. For the tokens to be added, the enroll information is
    provided containing the tokens secret. For the tokens to be updated, the token details are returned as dictionary.
    Additionally, the container rights read from the policies are included in the response.
    Additional parameters and entries in the response are possible, depending on the container type.
    The container is only authorized to synchronize if the challenge is valid.

    :jsonparam container_serial: Serial of the container
    :jsonparam container_dict_client: container data with included tokens from the client. The provided information
        may differ for different container and token types. To identify tokens at least the serial shall be provided.
        However, some clients might not have the serial. In this case, the client can provide a list of at least two
        otp values for hotp, totp and daypassword tokens.

    An example container_dict_client looks like this:
        ::

            {
                "serial": "SMPH001",
                "type": "smartphone",
                "tokens": [{"serial": "TOTP001", ...},
                           {"otp": ["1234", "4567"], "tokentype": "hotp"}]
            }

    :return: dictionary including the container properties, the tokens and the container policies.
        The provided enroll information depends on the token type as well as the returned information for the tokens to
        be updated.

    Example response:
        ::

            {
                "container": {"type": "smartphone", "serial": "SMPH001"},
                    "tokens": {"add": ["enroll_url1", "enroll_url2"],
                               "update": [{"serial": "TOTP001", "tokentype": "totp"},
                                          {"serial": "HOTP001", "otp": ["1234", "9876"],
                                           "tokentype": "hotp", "counter": 2}]}
                "policies": {"container_client_rollover": True,
                             "initially_add_tokens_to_container": False,
                             "disable_client_token_deletion": True,
                             "disable_client_container_unregister": True}
            }

    """
    params = request.all_data
    container_serial = getParam(params, "container_serial", optional=False)
    container_client_str = getParam(params, "container_dict_client", optional=True)
    container_client = json.loads(container_client_str) if container_client_str else {}

    # write client token serials to audit log
    client_serials = ", ".join(
        [token.get("serial") or str(token.get("otp")) for token in container_client.get("tokens", [])])

    container = find_container_by_serial(container_serial)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "action_detail": f"client_serials={client_serials}"})

    # Get server url
    server_url = container.get_container_info_dict().get("server_url")
    if server_url is None:
        log.debug("Server url is not set in the container info. Ensure the container is registered correctly.")
        server_url = " "
    scope = create_endpoint_url(server_url, "container/synchronize")
    params.update({'scope': scope})

    # 2nd synchronization step: Validate challenge and get container diff between client and server
    container.check_challenge_response(params)
    initially_add_tokens = request.all_data.get("client_policies").get(ACTION.INITIALLY_ADD_TOKENS_TO_CONTAINER)
    container_dict = container.synchronize_container_details(container_client, initially_add_tokens)

    # Write token serials to audit log
    add_serials = ", ".join([serial for serial in container_dict["tokens"]["add"]])
    equal_serials = ", ".join([token.get("serial") for token in container_dict["tokens"]["update"]])
    audit_info = f"Client: add={add_serials}, update={equal_serials}"

    # Get enroll information for missing tokens
    enroll_info = []
    org_all_data = request.all_data
    for serial in container_dict["tokens"]["add"]:
        try:
            # token rollover + get enroll url
            request.all_data = org_all_data
            enroll_url = regenerate_enroll_url(serial, request, g)
            if enroll_url:
                enroll_info.append(enroll_url)
            else:
                log.debug(f"Could not regenerate the enroll url for the token {serial} during synchronization of "
                          f"container {container_serial}.")
        except Exception as e:
            log.error(f"Could not regenerate the enroll url for the token {serial} during synchronization of"
                      f"container {container_serial}: {e}")
    container_dict["tokens"]["add"] = enroll_info

    # Optionally encrypt dict
    res = container.encrypt_dict(container_dict, params)
    res.update({'server_url': server_url})

    # Add policy info
    res["policies"] = request.all_data.get("client_policies")

    # Update last sync time
    container.update_last_synchronization()

    # Rollover completed: Change registration state to registered
    registration_state = container.get_container_info_dict().get("registration_state")
    if registration_state == "rollover_completed":
        container.update_container_info(
            [TokenContainerInfoData(key="registration_state", value="registered", info_type=PI_INTERNAL)])
    audit_info += f", rollover: {registration_state == 'rollover'}"

    # Audit log
    g.audit_object.log({"info": audit_info,
                        "success": True})

    return send_result(res)


@container_blueprint.route('/rollover', methods=['POST'])
@prepolicy(check_client_container_action, request, action=ACTION.CONTAINER_CLIENT_ROLLOVER)
@prepolicy(container_registration_config, request)
@event('container_init_rollover', request, g)
def rollover():
    """
    Initiate a rollover for a container which will generate new token secrets for all tokens in the container.
    The data or QR code is returned for the container to re-register.
    This endpoint can be used to transfer a container from one device to another.
    Parameters and entries in the returned dictionary are container type specific.

    :jsonparam container_serial: Serial of the container
    :return: Result of the rollover process as dictionary. The information may differ depending on the container
        type.

    An example response for smartphones looks like this:
        ::

            {
                "container_url": {"description": "URL for privacyIDEA Container Registration",
                                  "img": <QR code>,
                                  "value": "pia://container/SMPH0006D5BC?issuer=privacyIDEA&ttl=10..."},
                "nonce": "c238392af49250804c25bbd7d86408839e91fe97",
                "time_stamp": "2024-12-20T09:53:40.158319+00:00",
                "server_url": "https://pi.net",
                "ttl": 10,
                "ssl_verify": "True",
                "key_algorithm": "secp384r1",
                "hash_algorithm": "sha256",
                "passphrase_prompt": ""
            }
    """
    params = request.all_data
    container_serial = getParam(params, "container_serial", optional=False)
    container = find_container_by_serial(container_serial)
    # Params set by pre-policies
    server_url = getParam(params, "server_url")
    challenge_ttl = getParam(params, "challenge_ttl")
    registration_ttl = getParam(params, "registration_ttl")
    ssl_verify = getParam(params, "ssl_verify")

    # Check registration state: rollover is only allowed for registered containers
    registration_state = container.get_container_info_dict().get("registration_state")
    if registration_state != "registered" and registration_state != "rollover":
        raise ContainerNotRegistered(f"Container is not registered.")

    # Rollover
    res_rollover = init_container_rollover(container, server_url, challenge_ttl, registration_ttl, ssl_verify, params)

    # Audit log
    info_str = (f"server_url={server_url}, challenge_ttl={challenge_ttl}min, registration_ttl={registration_ttl}min, "
                f"ssl_verify={ssl_verify}, registration_state={registration_state}")
    g.audit_object.log({"container_serial": container_serial,
                        "container_type": container.type,
                        "info": info_str,
                        "success": True})

    return send_result(res_rollover)


# TEMPLATES
@container_blueprint.route('/templates', methods=['GET'])
@prepolicy(check_base_action, request, action=ACTION.CONTAINER_TEMPLATE_LIST)
@log_with(log)
def get_template():
    """
    Get all container templates filtered by the given parameters.

    :query name: Name of the template, optional
    :query container_type: Type of the container, optional
    :query page: Number of the page (starts with 1), optional
    :query pagesize: Number of templates displayed per page, optional
    :query sortdir: Sort direction, optional, default is "asc"
    :query sortby: column name to sort by, optional, default is "name"

    :return: Dictionary with at least an entry "templates" and further entries if pagination is used.

    An example response looks like this:
    ::

        {
            "templates": [{"name": "template1", "container_type": "smartphone",
                           "template_options": {"tokens": [{"type": "hotp", "genkey": True}, ...]}, ...},
                           {"name": "template2", "container_type": "yubikey", ...},
                           ...],
            "count": 25,
            "current": 1,
            "prev": null,
            "next": 2,
        }
    """
    params = request.all_data
    name = getParam(params, "name", optional=True)
    container_type = getParam(params, "container_type", optional=True)
    page = int(getParam(params, "page", optional=True, default=0) or 0)
    pagesize = int(getParam(params, "pagesize", optional=True, default=0) or 0)
    sortdir = getParam(params, "sortdir", optional=True, default="asc")
    sortby = getParam(params, "sortby", optional=True, default="name")

    templates_dict = get_templates_by_query(name=name, container_type=container_type, page=page, pagesize=pagesize,
                                            sortdir=sortdir, sortby=sortby)

    # Audit log
    g.audit_object.log({"success": True})

    return send_result(templates_dict)


@container_blueprint.route('<string:container_type>/template/<string:template_name>', methods=['POST'])
@prepolicy(check_base_action, request, action=ACTION.CONTAINER_TEMPLATE_CREATE)
@log_with(log)
def create_template_with_name(container_type, template_name):
    """
    Creates a template for the given name. If a template with this name already exists, the template options will be
    updated.

    :param container_type: Type of the container
    :param template_name: Name of the template
    :jsonparam template_options: Dictionary with the template options
    :jsonparam default: Set this template as default for the container type
    :return: ID of the created template or the template that was updated as dictionary such as
        ::

            {
                "template_id": 1
            }
    """
    params = request.all_data
    template_options = getParam(params, "template_options", optional=True) or {}
    default_template = getParam(params, "default", optional=True, default=False)

    # Audit log
    g.audit_object.log({"container_type": container_type,
                        "action_detail": f"template_name={template_name}, default={default_template}"})

    # Check parameters
    if not isinstance(template_options, dict):
        raise ParameterError("'template_options' must be a dictionary!")

    # check if name already exists
    existing_templates = get_templates_by_query(template_name)["templates"]

    if len(existing_templates) > 0:
        # update existing template
        template = get_template_obj(template_name)
        template.template_options = template_options
        template_id = template.id
        log.info(f"A template with the name '{template_name}' already exists. Updating template options.")
        audit_info = "Updated existing template"
    else:
        # create new template
        template_id = create_container_template(container_type, template_name, template_options, default_template)
        audit_info = "Created new template"

    # Set template as default for this container type
    if default_template:
        set_default_template(template_name)

    # Audit log
    g.audit_object.log({"success": True,
                        "info": audit_info})
    return send_result({"template_id": template_id})


@container_blueprint.route('template/<string:template_name>', methods=['DELETE'])
@prepolicy(check_base_action, request, action=ACTION.CONTAINER_TEMPLATE_DELETE)
@log_with(log)
def delete_template(template_name):
    """
    Deletes the template of the given name.

    :return: True if the template was deleted successfully, raises an exception otherwise
    """
    # Audit log
    g.audit_object.log({"action_detail": f"template_name={template_name}"})

    template = get_template_obj(template_name)
    template.delete()

    # Audit log
    g.audit_object.log({"container_type": template.get_class_type(),
                        "success": True})

    return send_result(True)


@container_blueprint.route('template/<string:template_name>/compare', methods=['GET'])
@prepolicy(check_base_action, request, action=ACTION.CONTAINER_TEMPLATE_LIST)
@prepolicy(check_base_action, request, action=ACTION.CONTAINER_LIST)
@prepolicy(check_admin_tokenlist, request, action=ACTION.CONTAINER_LIST)
@log_with(log)
def compare_template_with_containers(template_name):
    """
    Compares a template with its created containers.
    Only containers the user is allowed to manage are included in the comparison.

    If a container serial is provided, only this container will be compared to the template.

    :param template_name: Name of the template
    :jsonparam container_serial: Serial of the container to compare with the template, optional
    :return: A dictionary with the differences between the template and each container in the format:

        ::

            {"SMPH0001": {
                            "tokens": {
                                        "missing": ["hotp"],
                                        "additional": ["totp"]
                                        }
                            }
            }
    """
    allowed_realms = getattr(request, "pi_allowed_container_realms", None)
    user = request.User
    user_role = g.logged_in_user.get("role")
    container_serial = getParam(request.all_data, "container_serial", optional=True)

    # Audit log
    g.audit_object.log({"container_serial": container_serial,
                        "action_detail": f"template_name={template_name}"})

    template = get_template_obj(template_name)
    if container_serial:
        container_list = [find_container_by_serial(container_serial)]
    else:
        container_list = [create_container_from_db_object(db_container) for db_container in template.containers]

    result = {}
    for container in container_list:
        authorized = False
        if user_role == "admin":
            if allowed_realms is None:
                authorized = True
            else:
                container_realms = get_container_realms(container.serial)
                matching_realms = list(set(container_realms).intersection(allowed_realms))
                authorized = len(matching_realms) > 0
        elif user_role == "user":
            container_owners = container.get_users()
            authorized = user in container_owners

        if authorized:
            result[container.serial] = compare_template_with_container(template, container)
        else:
            log.info(f"User {user} is not authorized to access the container {container.serial}.")

    # Audit log
    g.audit_object.log({"container_type": template.get_class_type(),
                        "info": f"allowed_realms={allowed_realms}",
                        "success": True})

    return send_result(result)


@container_blueprint.route('template/tokentypes', methods=['GET'])
@log_with(log)
def get_template_token_types():
    """
    Returns a dictionary with the template container types as keys and their description and supported token types as
    values.

    ::

        {
            <type>: { description: "Description", token_types: ["hotp", "totp", "push", "daypassword", "sms"] },
            <type>: { description: "Description", token_types: ["hotp", "totp", "push", "daypassword", "sms"] }
        }

    """
    token_types = {}
    template_classes = get_container_template_classes()
    descriptions = get_container_classes_descriptions()

    for container_type in template_classes:
        token_types[container_type] = {"description": descriptions[container_type],
                                       "token_types": template_classes[container_type].get_supported_token_types()}

    g.audit_object.log({"success": True})
    return send_result(token_types)



# http://www.privacyidea.org
#
# 2018-08-01 Cornelius Kölbel, <cornelius.koelbel@netknights.it>
#            Initial writeup
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This endpoint is used fetch monitoring/statistics data

The code of this module is tested in tests/test_api_monitoring.py
"""
from flask import (Blueprint, request)
from privacyidea.api.lib.utils import getParam, send_result
from privacyidea.api.lib.prepolicy import prepolicy, check_base_action
from privacyidea.lib.utils import parse_legacy_time
from privacyidea.lib.log import log_with
from privacyidea.lib.monitoringstats import (get_stats_keys, get_values,
                                   get_last_value, delete_stats)
from privacyidea.lib.tokenclass import AUTH_DATE_FORMAT
from flask import g
import logging
from privacyidea.lib.policy import ACTION


log = logging.getLogger(__name__)


monitoring_blueprint = Blueprint('monitoring_blueprint', __name__)


@monitoring_blueprint.route('/', methods=['GET'])
@monitoring_blueprint.route('/<stats_key>', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.STATISTICSREAD)
def get_statistics(stats_key=None):
    """
    return a list of all available statistics keys in the database if no *stats_key*
    is specified.

    If a stats_key is specified it returns the data of this key.
    The parameters "start" and "end" can be used to specify a time window,
    from which the statistics data should be fetched.
    """
    if stats_key is None:
        stats_keys = get_stats_keys()
        g.audit_object.log({"success": True})
        return send_result(stats_keys)
    else:
        param = request.all_data
        start = getParam(param, "start")
        if start:
            start = parse_legacy_time(start, return_date=True)
        end = getParam(param, "end")
        if end:
            end = parse_legacy_time(end, return_date=True)
        values = get_values(stats_key=stats_key, start_timestamp=start, end_timestamp=end)
        # convert timestamps to strings
        values_w_string = [(s[0].strftime(AUTH_DATE_FORMAT), s[1]) for s in values]
        g.audit_object.log({"success": True})
        return send_result(values_w_string)


@monitoring_blueprint.route('/<stats_key>', methods=['DELETE'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.STATISTICSDELETE)
def delete_statistics(stats_key):
    """
    Delete the statistics data of a certain stats_key.

    You can specify the start date and the end date when to delete the
    monitoring data.
    You should specify the dates including the timezone. Otherwise your client
    could send its local time and the server would interpret it as its own local
    time which would result in deleting unexpected entries.

    You can specify the dates like 2010-12-31 22:00+0200
    """
    param = request.all_data
    start = getParam(param, "start")
    if start:
        start = parse_legacy_time(start, return_date=True)
    end = getParam(param, "end")
    if end:
        end = parse_legacy_time(end, return_date=True)
    r = delete_stats(stats_key, start, end)
    g.audit_object.log({"success": True})
    return send_result(r)


@monitoring_blueprint.route('/<stats_key>/last', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.STATISTICSREAD)
def get_statistics_last(stats_key):
    """
    Get the last value of the stats key
    """
    last_value = get_last_value(stats_key)
    g.audit_object.log({"success": True})
    return send_result(last_value)




"""
This test file tests the lib.resolvers.py

The lib.resolvers.py only depends on the database model.
"""
import uuid

from privacyidea.models import NodeName
from .base import MyTestCase

from privacyidea.lib.resolver import (save_resolver,
                                      delete_resolver)

from privacyidea.lib.realm import (set_realm,
                                   get_realms,
                                   get_default_realm,
                                   realm_is_defined,
                                   set_default_realm,
                                   delete_realm, export_realms, import_realms)


class ResolverTestCase(MyTestCase):
    """
    Test the token on the database level
    """
    resolvername1 = "resolver1"
    resolvername2 = "Resolver2"
    realm1 = "realm1"
    realm_dot = "realm1.com"

    def test_01_create_realm(self):
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": "/etc/passwd"})
        self.assertTrue(rid > 0, rid)

        rid = save_resolver({"resolver": self.resolvername2,
                             "type": "passwdresolver",
                             "fileName": "/etc/secrets"})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1,
                                    [{"name": self.resolvername1},
                                     {"name": self.resolvername2}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 2)

        (added, failed) = set_realm(self.realm_dot,
                                    [{"name": self.resolvername1},
                                     {"name": self.resolvername2}])
        self.assertEqual(0, len(failed))
        self.assertEqual(2, len(added))

        # test the realms
        realms = get_realms()
        self.assertTrue(self.realm1 in realms, realms)
        self.assertTrue(realms.get("realm1").get("default"), realms)
        self.assertTrue(self.realm_dot in realms, realms)
        # check that no node is set
        self.assertEqual("", realms.get("realm1").get("resolver")[0].get("node"), realms)
        self.assertEqual("", realms.get("realm1").get("resolver")[1].get("node"), realms)

        # delete dot realm
        delete_realm(self.realm_dot)

        # try to create realm with invalid name
        self.assertRaises(Exception, set_realm, "#####")

        # update the resolver list:
        (added, failed) = set_realm(self.realm1,
                                    [{'name': self.resolvername1},
                                     {'name': "non exiting"}])
        self.assertTrue(len(failed) == 1)
        self.assertTrue(len(added) == 1)

        self.assertTrue(realm_is_defined(self.realm1))
        self.assertTrue(realm_is_defined("non exist") is False)

    def test_03_get_specific_realm(self):
        realm = get_realms(self.realm1)
        self.assertTrue(self.realm1 in realm, realm)
        self.assertTrue(len(realm) == 1, realm)

    def test_02_set_default_realm(self):
        (added, failed) = set_realm("realm2",
                                    [{'name': self.resolvername2}])
        self.assertTrue(len(added) == 1)
        self.assertTrue(len(failed) == 0)

        realm = get_default_realm()
        self.assertTrue(realm == self.realm1)

        set_default_realm("realm2")
        realm = get_default_realm()
        self.assertTrue(realm == "realm2")

        set_default_realm()
        realm = get_default_realm()
        self.assertTrue(realm is None, realm)

    def test_10_delete_realm(self):
        delete_realm(self.realm1)
        delete_realm("realm2")
        delete_resolver(self.resolvername1)
        delete_resolver(self.resolvername2)
        realms = get_realms()
        self.assertTrue(len(realms) == 0, realms)

    def test_20_realms_with_nodes(self):
        nd1_uuid = "8e4272a9-9037-40df-8aa3-976e4a04b5a9"
        nd2_uuid = "d1d7fde6-330f-4c12-88f3-58a1752594bf"
        NodeName(id=nd1_uuid, name="Node1").save()
        NodeName(id=nd2_uuid, name="Node2").save()

        save_resolver({"resolver": self.resolvername1,
                       "type": "passwdresolver",
                       "fileName": "/etc/passwd"})

        save_resolver({"resolver": self.resolvername2,
                       "type": "passwdresolver",
                       "fileName": "/etc/secrets"})

        (added, failed) = set_realm("realm1",
                                    [
                                        {'name': self.resolvername1,
                                         'node': uuid.UUID(nd1_uuid)}])
        self.assertIn(self.resolvername1, added, added)
        self.assertEqual(len(failed), 0, failed)
        realms_dict = get_realms()
        self.assertIn("realm1", realms_dict, realms_dict)
        resolver_list = realms_dict["realm1"]["resolver"]
        self.assertEqual(len(resolver_list), 1, realms_dict)
        self.assertEqual(resolver_list[0]["node"], nd1_uuid, resolver_list)

        # overwrite/update existing realm with resolvers with different nodes
        (added, failed) = set_realm("realm1",
                                    [
                                        {'name': self.resolvername1,
                                         'node': uuid.UUID(nd1_uuid)},
                                        {'name': self.resolvername1,
                                         'node': uuid.UUID(nd2_uuid)}])
        self.assertIn(self.resolvername1, added, added)
        self.assertEqual(len(failed), 0, failed)
        realms_dict = get_realms()
        self.assertIn("realm1", realms_dict, realms_dict)
        resolver_list = realms_dict["realm1"]["resolver"]
        self.assertEqual(len(resolver_list), 2, realms_dict)
        self.assertIn(nd1_uuid, [x.get('node') for x in resolver_list], resolver_list)
        self.assertIn(nd2_uuid, [x.get('node') for x in resolver_list], resolver_list)

        # different resolvers, one with node, one without
        (added, failed) = set_realm("realm1",
                                    [
                                        {'name': self.resolvername1,
                                         'node': uuid.UUID(nd1_uuid)},
                                        {'name': self.resolvername2}])
        self.assertIn(self.resolvername1, added, added)
        self.assertIn(self.resolvername2, added, added)
        self.assertEqual(len(failed), 0, failed)
        realms_dict = get_realms()
        self.assertIn("realm1", realms_dict, realms_dict)
        resolver_list = realms_dict["realm1"]["resolver"]
        self.assertEqual(len(resolver_list), 2, realms_dict)
        self.assertEqual(
            nd1_uuid,
            next(x.get('node') for x in resolver_list if x.get("name") == self.resolvername1),
            resolver_list)
        self.assertEqual(
            "",
            next(x.get('node') for x in resolver_list if x.get("name") == self.resolvername2),
            resolver_list)

        # different resolvers, different nodes
        (added, failed) = set_realm("realm1",
                                    [
                                        {'name': self.resolvername1,
                                         'node': uuid.UUID(nd1_uuid)},
                                        {'name': self.resolvername2,
                                         'node': uuid.UUID(nd2_uuid)}])
        self.assertIn(self.resolvername1, added, added)
        self.assertIn(self.resolvername2, added, added)
        self.assertEqual(len(failed), 0, failed)
        realms_dict = get_realms()
        self.assertIn("realm1", realms_dict, realms_dict)
        resolver_list = realms_dict["realm1"]["resolver"]
        self.assertEqual(len(resolver_list), 2, realms_dict)
        self.assertEqual(
            nd1_uuid,
            next(x.get('node') for x in resolver_list if x.get("name") == self.resolvername1),
            resolver_list)
        self.assertEqual(
            nd2_uuid,
            next(x.get('node') for x in resolver_list if x.get("name") == self.resolvername2),
            resolver_list)

        delete_realm('realm1')
        NodeName.query.filter_by(id=nd1_uuid).delete()
        NodeName.query.filter_by(id=nd2_uuid).delete()

    def test_30_realm_import_export(self):
        nd1_uuid = "8e4272a9-9037-40df-8aa3-976e4a04b5a9"
        nd2_uuid = "d1d7fde6-330f-4c12-88f3-58a1752594bf"
        NodeName(id=nd1_uuid, name="Node1").save()
        NodeName(id=nd2_uuid, name="Node2").save()

        save_resolver({"resolver": self.resolvername1,
                       "type": "passwdresolver",
                       "fileName": "/etc/passwd"})

        save_resolver({"resolver": self.resolvername2,
                       "type": "passwdresolver",
                       "fileName": "/etc/secrets"})

        set_realm("realm1",
                  [
                      {'name': self.resolvername1,
                       'node': uuid.UUID(nd1_uuid)},
                      {'name': self.resolvername1}])
        realm_exp = export_realms("realm1")
        self.assertIn("realm1", realm_exp, realm_exp)
        self.assertTrue(realm_exp["realm1"]["default"], realm_exp)
        self.assertIn(self.resolvername1, [x["name"] for x in realm_exp["realm1"]["resolver"]], realm_exp)
        self.assertIn(nd1_uuid, [x["node"] for x in realm_exp["realm1"]["resolver"]], realm_exp)

        import_dict = {
            'realm2': {'id': 2, 'option': '', 'default': False, 'resolver': [
                {'priority': None, 'name': 'Resolver2', 'type': 'passwdresolver',
                 'node': nd2_uuid},
                {'priority': None, 'name': 'resolver1', 'type': 'passwdresolver',
                 'node': ''}]}}

        import_realms(import_dict)
        realms_dict = get_realms()
        self.assertEqual(len(realms_dict), 2, realms_dict)
        self.assertIn("realm2", realms_dict, realms_dict)
        resolver_list = realms_dict["realm2"]["resolver"]
        self.assertEqual(len(resolver_list), 2, realms_dict)
        self.assertIn(nd2_uuid, [x.get('node') for x in resolver_list], resolver_list)
        self.assertIn(self.resolvername2, [x.get('name') for x in resolver_list], resolver_list)

        delete_realm('realm1')
        delete_realm('realm2')
        delete_resolver("resolver1")
        delete_resolver("Resolver2")
        NodeName.query.filter_by(id=nd1_uuid).delete()
        NodeName.query.filter_by(id=nd2_uuid).delete()



#  2018-06-20 Friedrich Weber <friedrich.weber@netknights.it>
#             Initial skeleton
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the base class for a task module.

Tasks may be run periodically. Upon execution, a task
can be given a set of options.
"""


class BaseTask(object):
    """
    A BaseTask returns a list of supported options.
    """

    identifier = "BaseTask"
    description = "This is the base class of a task with no functionality"

    def __init__(self, config):
        self.config = config

    @property
    def options(self):
        """
        A task may be given a dictionary of options. The allowed keys, their
        descriptions and types are specified here.

        For each option key, the returned dictionary contains one nested dictionary.
        Each nested dictionary has keys "type", "desc" and "value".

        :return: dict
        """
        return {}

    def do(self, params=None):
        """
        This method executes the task with the given parameters.

        :param params: a dictionary
        :return: a boolean denoting success
        """
        return True



# http://www.privacyidea.org
# (c) Cornelius Kölbel
#
# 2015-12-27 Cornelius Kölbel, <cornelius@privacyidea.org>
#            Implement REST API, create, update, delete, list
#            for SMTP server definitions
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This endpoint is used to create, update, list and delete SMTP
server definitions. SMTP server definitions can be used for several purposes
like
EMail-Token, SMS Token with SMTP gateway, notification like PIN handler and
registration.

The code of this module is tested in tests/test_api_smtpserver.py
"""
from flask import (Blueprint,
                   request, current_app)
from .lib.utils import (getParam,
                        required,
                        send_result)
from ..lib.log import log_with
from ..lib.crypto import decryptPassword, FAILED_TO_DECRYPT_PASSWORD
from ..lib.policy import ACTION
from ..lib.utils import is_true
from ..api.lib.prepolicy import prepolicy, check_base_action
from flask import g
from flask_babel import gettext as _
import logging
from privacyidea.lib.smtpserver import (add_smtpserver, list_smtpservers,
                                        delete_smtpserver, send_or_enqueue_email)

log = logging.getLogger(__name__)

smtpserver_blueprint = Blueprint('smtpserver_blueprint', __name__)


@smtpserver_blueprint.route('/<identifier>', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.SMTPSERVERWRITE)
@log_with(log)
def create(identifier=None):
    """
    This call creates or updates an SMTP server definition.

    :param identifier: The unique name of the SMTP server definition
    :param server: The FQDN or IP of the mail server
    :param port: The port of the mail server
    :param username: The mail username for authentication at the SMTP server
    :param password: The password for authentication at the SMTP server
    :param tls: If the server should do TLS
    :param description: A description for the definition
    """
    param = request.all_data
    server = getParam(param, "server", required)
    port = int(getParam(param, "port", default=25))
    username = getParam(param, "username", default="")
    password = getParam(param, "password", default="")
    sender = getParam(param, "sender", default="")
    tls = is_true(getParam(param, "tls", default=False))
    description = getParam(param, "description", default="")
    timeout = int(getParam(param, "timeout") or 10)
    enqueue_job = is_true(getParam(param, "enqueue_job", default=False))

    r = add_smtpserver(identifier, server, port=port, username=username,
                       password=password, tls=tls, description=description,
                       sender=sender, timeout=timeout, enqueue_job=enqueue_job)

    g.audit_object.log({'success': r > 0,
                        'info':  r})
    return send_result(r > 0)


@smtpserver_blueprint.route('/', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.SMTPSERVERREAD)
def list_smtpservers_api():
    """
    This call gets the list of SMTP server definitions
    """
    res = list_smtpservers()
    g.audit_object.log({'success': True})
    return send_result(res)


@smtpserver_blueprint.route('/<identifier>', methods=['DELETE'])
@prepolicy(check_base_action, request, ACTION.SMTPSERVERWRITE)
@log_with(log)
def delete_server(identifier=None):
    """
    This call deletes the specified SMTP server configuration

    :param identifier: The unique name of the SMTP server definition
    """
    r = delete_smtpserver(identifier)

    g.audit_object.log({'success': r > 0,
                        'info':  r})
    return send_result(r > 0)


@smtpserver_blueprint.route('/send_test_email', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.SMTPSERVERWRITE)
@log_with(log)
def test():
    """
    Test the email configuration
    :return:
    """
    param = request.all_data
    identifier = getParam(param, "identifier", required)
    server = getParam(param, "server", required)
    port = int(getParam(param, "port", default=25))
    username = getParam(param, "username", default="")
    password = getParam(param, "password", default="")
    sender = getParam(param, "sender", default="")
    tls = is_true(getParam(param, "tls", default=False))
    recipient = getParam(param, "recipient", required)
    timeout = int(getParam(param, "timeout") or 10)
    enqueue_job = is_true(getParam(param, "enqueue_job", default=False))

    s = dict(identifier=identifier, server=server, port=port,
             username=username, password=password, sender=sender,
             tls=tls, timeout=timeout, enqueue_job=enqueue_job)
    r = send_or_enqueue_email(s, recipient,
                              "Test Email from privacyIDEA",
                              "This is a test email from privacyIDEA. "
                              "The configuration {} is working.".format(identifier))

    g.audit_object.log({'success': r > 0,
                        'info':  r})
    return send_result(r > 0)



"""Add column for triggered policies in Audit Table.

Revision ID: a63df077051a
Revises: d2ae8e54b628
Create Date: 2018-10-31 16:00:49.219821

"""

# revision identifiers, used by Alembic.
revision = 'a63df077051a'
down_revision = 'd2ae8e54b628'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('pidea_audit', sa.Column('policies', sa.String(length=255), nullable=True))
    except Exception as exx:
        print('Adding of column "policies" in table pidea_audit failed: {!r}'.format(exx))
        print('This is expected behavior if this column already exists.')


def downgrade():
    op.drop_column('pidea_audit', 'policies')




"""
This test file tests the lib.tokens.sshkeytoken
This depends on lib.tokenclass
"""
from privacyidea.lib.error import TokenAdminError
from .base import MyTestCase
from privacyidea.lib.tokenclass import ROLLOUTSTATE
from privacyidea.lib.tokens.sshkeytoken import SSHkeyTokenClass
from privacyidea.models import Token


class SSHTokenTestCase(MyTestCase):

    otppin = "topsecret"
    serial1 = "ser1"
    serial2 = "ser2"
    serial3 = "ser3"
    serial4 = "ser4"
    sshkey = "ssh-rsa " \
             "AAAAB3NzaC1yc2EAAAADAQABAAACAQDJy0rLoxqc8SsY8DVAFijMsQyCv" \
             "hBu4K40hdZOacXK4O6OgnacnSKN56MP6pzz2+4svzvDzwvkFsvf34pbsgD" \
             "F67PPSCsimmjEQjf0UfamBKh0cl181CbPYsph3UTBOCgHh3FFDXBduPK4DQz" \
             "EVQpmqe80h+lsvQ81qPYagbRW6fpd0uWn9H7a/qiLQZsiKLL07HGB+NwWue4os" \
             "0r9s4qxeG76K6QM7nZKyC0KRAz7CjAf+0X7YzCOu2pzyxVdj/T+KArFcMmq8V" \
             "dz24mhcFFXTzU3wveas1A9rwamYWB+Spuohh/OrK3wDsrryStKQv7yofgnPMs" \
             "TdaL7XxyQVPCmh2jVl5ro9BPIjTXsre9EUxZYFVr3EIECRDNWy3xEnUHk7Rzs" \
             "734Rp6XxGSzcSLSju8/MBzUVe35iXfXDRcqTcoA0700pIb1ANYrPUO8Up05v4" \
             "EjIyBeU61b4ilJ3PNcEVld6FHwP3Z7F068ef4DXEC/d7pibrp4Up61WYQIXV/" \
             "utDt3NDg/Zf3iqoYcJNM/zIZx2j1kQQwqtnbGqxJMrL6LtClmeWteR4420uZx" \
             "afLE9AtAL4nnMPuubC87L0wJ88un9teza/N02KJMHy01Yz3iJKt3Ou9eV6kqO" \
             "ei3kvLs5dXmriTHp6g9whtnN6/Liv9SzZPJTs8YfThi34Wccrw== " \
             "NetKnights GmbH Descröption"
    unsupported_keytype = "ssh-something AAAAA comment"
    sshkey_ecdsa = "ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzd" \
                   "HAyNTYAAABBBHGCdIk0pO1HFr/mF4oLb43ZRyQJ4K7ICLrAhAiQERVa0tUvyY5TE" \
                   "zurWTqxSMx203rY77t6xnHLZBMPPpv8rk0= cornelius@puck"
    sshkey_ed25519 = "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIC38dIb3tM6nPrT" \
                     "3j1UfsQxOCBbf3JogwsKeVPM893Pi cornelius@puck"
    ecdsa_sk = "sk-ecdsa-sha2-nistp256@openssh.com AAAAInNrLWVjZHNhLXNoYTItbmlz" \
               "dHAyNTZAb3BlbnNzaC5jb20AAAAIbmlzdHAyNTYAAABBBOStamg+GO4TSgtoWjc82p" \
               "OKZIDuOeAt/8PU/jbzEmth6VuNhghRTCPqPMFtR6mB3Pb12yMDRiLH/t1VwkvWWYIA" \
               "AAAEc3NoOg=="
    wrong_sshkey = """---- BEGIN SSH2 PUBLIC KEY ----
    AAAAB3NzaC1kc3MAAACBAKrFC6uDvuxl9vnYL/Fu/Vq+12KJF4
    RyMSQe4mn8oHJma2VzepBRBpLt7Q==
    ---- END SSH2 PUBLIC KEY ----"""
    INVALID_SSH = "ssh-rsa"

    def test_01_create_token(self):
        db_token = Token(self.serial1, tokentype="sshkey")
        db_token.save()
        token = SSHkeyTokenClass(db_token)

        # An invalid key, raises an exception
        self.assertRaises(TokenAdminError, token.update, {"sshkey": "InvalidKey"})
        self.assertEqual(token.rollout_state, ROLLOUTSTATE.BROKEN)

        # An invalid key, raises an exception
        self.assertRaises(TokenAdminError, token.update, {"sshkey": self.INVALID_SSH})
        self.assertEqual(token.rollout_state, ROLLOUTSTATE.BROKEN)

        # An invalid key, raises an exception
        self.assertRaises(TokenAdminError, token.update, {"sshkey": self.wrong_sshkey})
        self.assertEqual(token.rollout_state, ROLLOUTSTATE.BROKEN)

        # An unsupported keytype
        self.assertRaises(TokenAdminError, token.update, {"sshkey": self.unsupported_keytype})
        self.assertEqual(token.rollout_state, ROLLOUTSTATE.BROKEN)

        # Set valid key
        token.update({"sshkey": self.sshkey})
        self.assertTrue(token.token.serial == self.serial1, token)
        self.assertTrue(token.token.tokentype == "sshkey",
                        token.token.tokentype)
        self.assertTrue(token.type == "sshkey", token)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "SSHK", class_prefix)
        self.assertTrue(token.get_class_type() == "sshkey", token)

        # ecdsa
        db_token = Token(self.serial2, tokentype="sshkey")
        db_token.save()
        token = SSHkeyTokenClass(db_token)
        token.update({"sshkey": self.sshkey_ecdsa})

        # ed25519
        db_token = Token(self.serial3, tokentype="sshkey")
        db_token.save()
        token = SSHkeyTokenClass(db_token)
        token.update({"sshkey": self.sshkey_ed25519})

        # ecdsa_sk
        db_token = Token(self.serial4, tokentype="sshkey")
        db_token.save()
        token = SSHkeyTokenClass(db_token)
        token.update({"sshkey": self.ecdsa_sk})

    def test_02_class_methods(self):
        db_token = Token.query.filter(Token.serial == self.serial1).first()
        token = SSHkeyTokenClass(db_token)

        info = token.get_class_info()
        self.assertTrue(info.get("title") == "SSHkey Token",
                        "{0!s}".format(info.get("title")))

        info = token.get_class_info("title")
        self.assertTrue(info == "SSHkey Token", info)

    def test_03_get_sshkey(self):
        db_token = Token.query.filter(Token.serial == self.serial1).first()
        token = SSHkeyTokenClass(db_token)
        sshkey = token.get_sshkey()
        self.assertTrue(sshkey == self.sshkey, sshkey)
        self.assertIsInstance(sshkey, str)

        db_token = Token.query.filter(Token.serial == self.serial2).first()
        token = SSHkeyTokenClass(db_token)
        sshkey = token.get_sshkey()
        self.assertTrue(sshkey == self.sshkey_ecdsa, sshkey)
        self.assertIsInstance(sshkey, str)

        db_token = Token.query.filter(Token.serial == self.serial3).first()
        token = SSHkeyTokenClass(db_token)
        sshkey = token.get_sshkey()
        self.assertTrue(sshkey == self.sshkey_ed25519, sshkey)
        self.assertIsInstance(sshkey, str)

        db_token = Token.query.filter(Token.serial == self.serial4).first()
        token = SSHkeyTokenClass(db_token)
        sshkey = token.get_sshkey()
        self.assertEqual(self.ecdsa_sk, sshkey)
        self.assertIsInstance(sshkey, str)



# (c) NetKnights GmbH 2025,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2025 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later

from dataclasses import dataclass

PI_INTERNAL = "pi_internal"

@dataclass
class TokenContainerInfoData:
    """
    Dataclass for token container info
    """
    key: str
    value: str
    type: str
    description: str

    def __init__(self, key: str, value: str, info_type: str = None, description: str = None):
        self.key = key
        self.value = value
        self.type = info_type
        self.description = description



"""
This file contains the tests for the lifecycle module lib/lifecycle.py
"""
from mock import mock

from privacyidea.lib.lifecycle import register_finalizer, call_finalizers
from .base import MyTestCase


class LifecycleTestCase(MyTestCase):
    # Create admin authentication token for each testcase
    def setUp(self):
        self.authenticate()

    def test_01_register_finalizer(self):
        finalizer1 = mock.MagicMock()
        finalizer2 = mock.MagicMock()
        finalizer3 = mock.MagicMock()
        register_finalizer(finalizer1)
        register_finalizer(finalizer2)
        call_finalizers()
        finalizer1.assert_called_once()
        finalizer2.assert_called_once()
        finalizer3.assert_not_called()
        # call_finalizer clears the list of finalizers
        register_finalizer(finalizer3)
        call_finalizers()
        finalizer1.assert_called_once()
        finalizer2.assert_called_once()
        finalizer3.assert_called_once()

    def test_02_register_finalizer_request_context(self):
        finalizer1 = mock.MagicMock()
        finalizer2 = mock.MagicMock()
        # test that we can use finalizers
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            register_finalizer(finalizer1)
            register_finalizer(finalizer2)
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertTrue(result.get("status"))
        finalizer1.assert_called_once()
        finalizer2.assert_called_once()
        # test that they are not called in the next request
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertTrue(result.get("status"))
        finalizer1.assert_called_once()
        finalizer2.assert_called_once()

    def test_03_finalizer_error(self):
        finalizer1 = mock.MagicMock()
        finalizer1.side_effect = RuntimeError()
        finalizer2 = mock.MagicMock()
        # test that we can use finalizers
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            register_finalizer(finalizer1)
            register_finalizer(finalizer2)
            res = self.app.full_dispatch_request()
            result = res.json.get("result")
            self.assertTrue(result.get("status"))
        finalizer1.assert_called_once()
        finalizer2.assert_called_once()



from .base import MyApiTestCase
from privacyidea.lib.monitoringstats import write_stats
from privacyidea.lib.tokenclass import AUTH_DATE_FORMAT
from privacyidea.models import db
import datetime


class APIMonitoringTestCase(MyApiTestCase):

    def test_01_get_stats(self):

        # create some statistics

        write_stats("key1", 1)
        write_stats("key2", 50)
        write_stats("key1", 2)
        ts = datetime.datetime.now().isoformat()

        write_stats("key2", 60)
        write_stats("key1", 3)
        write_stats("key2", 50)
        write_stats("key1", 4)

        # get available stats keys
        with self.app.test_request_context('/monitoring/',
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertIn("key1", result.get("value"), result)
            self.assertIn("key2", result.get("value"), result)

        # check values of key1
        with self.app.test_request_context('/monitoring/key1',
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(len(result.get("value")), 4)
            self.assertEqual(result.get("value")[0][1], 1)
            self.assertEqual(result.get("value")[1][1], 2)
            self.assertEqual(result.get("value")[2][1], 3)
            self.assertEqual(result.get("value")[3][1], 4)

        # check values of key1, with a start value in the past
        with self.app.test_request_context('/monitoring/key1',
                                           data={"start": "2010-01-01 10:00+0200"},
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(len(result.get("value")), 4)

        # End value in the past will return no data.
        with self.app.test_request_context('/monitoring/key1',
                                           query_string={"end": "2010-01-01"},
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(len(result.get("value")), 0)

        # check with start timestamp after the 2nd value.
        # This should return the 3rd and 4th.
        with self.app.test_request_context('/monitoring/key1',
                                           query_string={"start": ts},
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(len(result.get("value")), 2)
            self.assertEqual(result.get("value")[0][1], 3)
            self.assertEqual(result.get("value")[1][1], 4)

        # check the last value of key1
        with self.app.test_request_context('/monitoring/key1/last',
                                           method='GET',
                                           headers={'Authorization': self.at}):

            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(4, result.get("value"), result)

    def test_02_delete_stats(self):

        ts = datetime.datetime.now()
        write_stats("key2", 60)

        # Now we delete some keys (the three old ones)
        with self.app.test_request_context('/monitoring/key2',
                                           method='DELETE',
                                           data={"start": "2010-01-01",
                                                 "end": ts.strftime(AUTH_DATE_FORMAT)},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            # Number of deleted values
            self.assertEqual(result.get("value"), 3)

        # ..and check if there is only one key left!
        with self.app.test_request_context('/monitoring/key2',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            # Number of remaining values
            self.assertEqual(1, len(result.get("value")), result)
